"""Unit test package for xbitinfo."""

import importlib

import pytest
from packaging.version import InvalidVersion, Version


def _import_or_skip(modname, minversion=None):
    try:
        mod = importlib.import_module(modname)
        has = True
        if minversion is not None:
            try:
                if Version(mod.__version__) < Version(minversion):
                    raise ImportError("Minimum version not satisfied")
            except InvalidVersion:
                raise ImportError(f"Invalid version for {modname}: {mod.__version__}")
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


has_julia, requires_julia = _import_or_skip("julia")



import xarray as xr
from numcodecs.bitround import BitRound

from .xbitinfo import _jl_bitround, get_keepbits


def _np_bitround(data, keepbits):
    """Bitround for Arrays."""
    codec = BitRound(keepbits=keepbits)
    data = data.copy()  # otherwise overwrites the input
    encoded = codec.encode(data)
    return codec.decode(encoded)


def _keepbits_interface(da, keepbits):
    """Common interface to allowed keepbits types

    Parameters
    ----------
    da : :py:class:`xarray.DataArray`
      Input data to bitround
    keepbits : int, dict of {str: int}, :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`
      How many bits to keep as int

    Returns
    -------
    keep : int
      Number of keepbits for variable given in ``da``
    """
    assert isinstance(da, xr.DataArray)
    if isinstance(keepbits, int):
        keep = keepbits
    elif isinstance(keepbits, dict):
        v = da.name
        if v in keepbits.keys():
            keep = keepbits[v]
        else:
            raise ValueError(f"name {v} not for in keepbits: {keepbits.keys()}")
    elif isinstance(keepbits, xr.Dataset):
        assert keepbits.coords["inflevel"].shape <= (
            1,
        ), "Information content is only allowed for one 'inflevel' here. Please make a selection."
        if "dim" in keepbits.coords:
            assert keepbits.coords["dim"].shape <= (
                1,
            ), "Information content is only allowed along one dimension here. Please select one `dim`. To find the maximum keepbits, simply use `keepbits.max(dim='dim')`"
        v = da.name
        if v in keepbits.keys():
            keep = int(keepbits[v])
        else:
            raise ValueError(f"name {v} not for in keepbits: {keepbits.keys()}")
    elif isinstance(keepbits, xr.DataArray):
        assert keepbits.coords["inflevel"].shape <= (
            1,
        ), "Information content is only allowed for one 'inflevel' here. Please make a selection."
        assert keepbits.coords["dim"].shape <= (
            1,
        ), "Information content is only allowed along one dimension here. Please select one `dim`. To find the maximum keepbits, simply use `keepbits.max(dim='dim')`"
        v = da.name
        if v == keepbits.name:
            keep = int(keepbits)
        else:
            raise KeyError(f"no keepbits found for variable {v}")
    else:
        raise TypeError(f"type {type(keepbits)} is not a valid type for keepbits.")
    return keep


def xr_bitround(da, keepbits):
    """Apply bitrounding based on keepbits from :py:func:`xbitinfo.xbitinfo.get_keepbits` for :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray` wrapping ``numcodecs.bitround``

    Parameters
    ----------
    da : :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`
      Input data to bitround
    keepbits : int, dict of {str: int}, :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`
      How many bits to keep as int. Fails if dict or :py:class:`xarray.Dataset` and key or variable not present.

    Returns
    -------
    da_bitrounded : :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info_per_bit = xb.get_bitinformation(ds, dim="lon")
    >>> keepbits = xb.get_keepbits(info_per_bit, 0.99)
    >>> ds_bitrounded = xb.xr_bitround(ds, keepbits)
    """
    if isinstance(da, xr.Dataset):
        da_bitrounded = da.copy()
        for v in da.data_vars:
            da_bitrounded[v] = xr_bitround(da[v], keepbits)
        return da_bitrounded

    assert isinstance(da, xr.DataArray)
    keep = _keepbits_interface(da, keepbits)

    da = xr.apply_ufunc(_np_bitround, da, keep, dask="parallelized", keep_attrs=True)
    da.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] = keep
    return da


def jl_bitround(da, keepbits):
    """Apply bitrounding based on keepbits from :py:func:`xbitinfo.xbitinfo.get_keepbits` for :py:class:`xarray.Dataset` or :py:class:`xarray.DataArray` wrapping `BitInformation.jl.round <https://github.com/milankl/BitInformation.jl/blob/main/src/round_nearest.jl>`__.

    Parameters
    ----------
    da : :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`
      Input data to bitround
    keepbits : int, dict of {str: int}, :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`
      How many bits to keep as int. Fails if dict or :py:class:`xarray.Dataset` and key or variable not present.

    Returns
    -------
    da_bitrounded : :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info_per_bit = xb.get_bitinformation(ds, dim="lon")
    >>> keepbits = xb.get_keepbits(info_per_bit, 0.99)
    >>> ds_bitrounded = xb.jl_bitround(ds, keepbits)
    """
    if isinstance(da, xr.Dataset):
        da_bitrounded = da.copy()
        for v in da.data_vars:
            da_bitrounded[v] = jl_bitround(da[v], keepbits)
        return da_bitrounded

    assert isinstance(da, xr.DataArray)
    keep = _keepbits_interface(da, keepbits)
    da = xr.apply_ufunc(_jl_bitround, da, keep, dask="forbidden", keep_attrs=True)
    da.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] = keep
    return da


def bitround_along_dim(
    ds, info_per_bit, dim, inflevels=[1.0, 0.9999, 0.99, 0.975, 0.95], keepbits=None
):
    """
    Apply bitrounding on slices along dim based on inflevels.
    Helper function to generate data for Fig. 3 in Klöwer et al. 2021.

    Klöwer, M., Razinger, M., Dominguez, J. J., Düben, P. D., & Palmer, T. N. (2021).
    Compressing atmospheric data into its real information content.
    Nature Computational Science, 1(11), 713–724. doi: 10/gnm4jj

    Parameters
    ----------
    ds : :py:class:`xarray.Dataset`, :py:class:`xarray.DataArray`
      Input
    info_per_bit : dict
      Information content of each bit for each variable in ds. This is the output from get_bitinformation.
    dim : str
      Name of dimension for slicing
    inflevels : list of floats
      Level of information that shall be preserved. Defaults to ``[1.0, 0.9999, 0.99, 0.975, 0.95]``.
    keepbits : int, dict of {str: int}, :py:class:`xarray.DataArray` or :py:class:`xarray.Dataset`
      How many bits to keep as int

    Returns
    -------
    ds : :py:class:`xarray.Dataset`, :py:class:`xarray.DataArray`
      Bitrounded on slices along ``dim`` based on ``inflevels``

    Raises
    ------
    ValueError
        If both `inflevels` and `keepbits` are specified, or if neither is specified.

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info_per_bit = xb.get_bitinformation(ds, dim="lon")
    >>> ds_bitrounded_along_lon = xb.bitround.bitround_along_dim(
    ...     ds, info_per_bit, dim="lon"
    ... )
    >>> (ds - ds_bitrounded_along_lon)["air"].isel(time=0).plot()  # doctest: +ELLIPSIS
    <matplotlib.collections.QuadMesh object at ...>

    """
    new_ds = []
    if inflevels is not None and keepbits is not None:
        raise ValueError("Either inflevel or keepbits should be None")
    elif inflevels is not None:
        stride = ds[dim].size // len(inflevels)
        for i, inf in enumerate(inflevels):  # last slice might be a bit larger
            ds_slice = ds.isel(
                {
                    dim: slice(
                        stride * i,
                        stride * (i + 1) if i != len(inflevels) - 1 else None,
                    )
                }
            )
            keepbits_slice = get_keepbits(info_per_bit, inf)
            if inf != 1:
                ds_slice_bitrounded = xr_bitround(ds_slice, keepbits_slice)
            else:
                ds_slice_bitrounded = ds_slice
            new_ds.append(ds_slice_bitrounded)
    elif keepbits is not None:
        new_ds = [xr_bitround(ds, keepbits)]
    else:
        raise ValueError("Either inflevel or keepbits should NOT be None")
    return xr.concat(new_ds, dim)



"""Top-level package for xbitinfo."""

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "unknown"

from .bitround import jl_bitround, xr_bitround
from .graphics import plot_bitinformation, plot_distribution
from .save_compressed import get_compress_encoding_nc, get_compress_encoding_zarr
from .xbitinfo import get_bitinformation, get_keepbits, get_prefect_flow



import os

import prefect
import pytest
import xarray as xr
from prefect_dask.task_runners import DaskTaskRunner

import xbitinfo as xb

if __name__ == "__main__":

    @pytest.mark.parametrize(
        "ds,dim,axis",
        [
            ("ugrid_demo", None, -1),
            ("icon_grid_demo", "ncells", None),
            ("air_temperature", "lon", None),
            ("rasm", "x", None),
            ("ROMS_example", "eta_rho", None),
            ("era52mt", "time", None),
            ("eraint_uvz", "longitude", None),
        ],
    )
    def test_full(ds, dim, axis, request):
        """Test xbitinfo end to end."""
        ds = request.getfixturevalue(ds)
        # xbitinfo
        bitinfo = xb.get_bitinformation(ds, dim=dim, axis=axis)
        keepbits = xb.get_keepbits(bitinfo)
        # ds_bitrounded = xb.jl_bitround(ds, keepbits)
        ds_bitrounded = xb.xr_bitround(ds, keepbits)  # identical
        # save
        label = os.path.basename(ds.encoding["source"])
        print(label)
        ds.to_netcdf(f"./tmp_testdir/{label}.nc")
        ds.to_compressed_netcdf(f"./tmp_testdir/{label}_compressed.nc")
        ds_bitrounded.to_compressed_netcdf(
            f"./tmp_testdir/{label}_bitrounded_compressed.nc"
        )
        # check size reduction
        ori_size = os.path.getsize(f"./tmp_testdir/{label}.nc")
        compressed_size = os.path.getsize(f"./tmp_testdir/{label}_compressed.nc")
        bitrounded_compressed_size = os.path.getsize(
            f"./tmp_testdir/{label}_bitrounded_compressed.nc"
        )
        if dim == "eta_rho":
            assert (
                compressed_size < ori_size * 1.1
            )  # previous compression is already really good for ROMS_example
        else:
            assert compressed_size < ori_size
        assert bitrounded_compressed_size < compressed_size

    def test_full_max_keepbits():
        """Test pipeline to get maximum keepbits"""
        label = "air_temperature"
        ds = xr.tutorial.load_dataset(label)
        bi = xb.get_bitinformation(ds)
        kb = xb.get_keepbits(bi)
        kb_max = kb.max(dim="dim")
        _ = xb.plot_bitinformation(bi.isel(dim=[0]))
        _ = xb.plot_bitinformation(bi.isel(dim=0))
        ds_bitrounded_max = xb.xr_bitround(ds, kb_max)
        ds_bitrounded_max.to_compressed_zarr(f"{label}.zarr", mode="w")

    imax = 3

    @pytest.fixture()
    def flow_paths(rasm):
        paths = []
        stride = rasm.time.size // imax
        for i in range(imax):
            f = f"./tmp_testdir/file_{i}.nc"
            if os.path.exists(f.replace(".nc", "_bitrounded_compressed.nc")):
                os.remove(f.replace(".nc", "_bitrounded_compressed.nc"))
            paths.append(f)
            rasm.isel(time=slice(stride * i, stride * (i + 1) - 1)).to_netcdf(f)
        flow = xb.get_prefect_flow(paths)
        yield flow, paths
        # cleanup
        for p in paths:
            if os.path.exists(p):
                os.remove(p)
            if os.path.exists(p.replace(".nc", "_bitrounded_compressed.nc")):
                os.remove(p.replace(".nc", "_bitrounded_compressed.nc"))

    @pytest.mark.parametrize(
        "executor",
        [
            "local",
            "my_dask_client",
            pytest.param(
                "DaskExecutor",
                marks=pytest.mark.skip(reason="fails with few resources"),
            ),
            pytest.param(
                "LocalDaskExecutor",
                marks=pytest.mark.skip(reason="fails with few resources"),
            ),
        ],
    )
    def test_get_prefect_flow_executor(flow_paths, executor):
        """Test get_prefect_flow runs for different executors."""
        flow, paths = flow_paths
        for f in paths:
            if os.path.exists(f.replace(".nc", "_bitrounded_compressed.nc")):
                os.remove(f.replace(".nc", "_bitrounded_compressed.nc"))

        if executor == "local":
            flow.run()
        elif executor == "my_dask_client":
            from dask.distributed import Client

            client = Client(n_workers=4, threads_per_worker=1, processes=True)
            # point Prefect's DaskTaskRunner to our Dask cluster
            dask_executor = DaskTaskRunner(address=client.scheduler.address)
            flow.run(executor=dask_executor)
            client.close()
        else:
            # For other executors, you can use the Prefect executor directly
            executor_class = getattr(prefect.executors, executor)
            flow.run(executor=executor_class())

    def test_get_prefect_flow_inflevel_parameter(flow_paths):
        """Test get_prefect_flow runs for different parameters."""
        flow, paths = flow_paths
        st090 = flow.run(parameters={"axis": -1, "inflevel": 0.90, "overwrite": True})
        st099999999 = flow.run(
            parameters={"axis": -1, "inflevel": 0.99999999, "overwrite": True}
        )
        keepbits = flow.get_tasks(name="get_bitinformation_keepbits")[0]

        assert st099999999.result[keepbits].result != st090.result[keepbits].result



import json
import logging
import os
import warnings

import numpy as np
import xarray as xr
from dask import array as da
from prefect import flow, task, unmapped

try:
    from julia.api import Julia

    julia_installed = True
except ImportError:
    julia_installed = False
from tqdm.auto import tqdm

import xbitinfo as xb

from . import _py_bitinfo as pb
from .julia_helpers import install

already_ran = False
if not already_ran and julia_installed:
    already_ran = install(quiet=True)
    jl = Julia(compiled_modules=False, debug=False)
    from julia import Main  # noqa: E402

    path_to_julia_functions = os.path.join(
        os.path.dirname(__file__), "bitinformation_wrapper.jl"
    )
    Main.path = path_to_julia_functions
    jl.using("BitInformation")
    jl.using("Pkg")
    jl.eval("include(Main.path)")


def bit_partitioning(dtype):
    if dtype.kind == "f":
        n_bits = np.finfo(dtype).bits
        n_sign = 1
        n_exponent = np.finfo(dtype).nexp
        n_mantissa = np.finfo(dtype).nmant
    elif dtype.kind == "i":
        n_bits = np.iinfo(dtype).bits
        n_sign = 1
        n_exponent = 0
        n_mantissa = n_bits - n_sign
    elif dtype.kind == "u":
        n_bits = np.iinfo(dtype).bits
        n_sign = 0
        n_exponent = 0
        n_mantissa = n_bits - n_sign
    else:
        raise ValueError(f"dtype {dtype} neither known nor implemented.")
    assert (
        n_sign + n_exponent + n_mantissa == n_bits
    ), "The components of the datatype could not be safely inferred."
    return n_bits, n_sign, n_exponent, n_mantissa


def get_bit_coords(dtype):
    """Get coordinates for bits based on dtype."""
    n_bits, n_sign, n_exponent, n_mantissa = bit_partitioning(dtype)
    coords = (
        n_sign * ["±"]
        + [f"e{int(i)}" for i in range(1, n_exponent + 1)]
        + [f"m{int(i)}" for i in range(1, n_mantissa + 1)]
    )
    return coords


def dict_to_dataset(info_per_bit):
    """Convert keepbits dictionary to :py:class:`xarray.Dataset`."""
    dsb = xr.Dataset()
    for v in info_per_bit.keys():
        dtype = np.dtype(info_per_bit[v]["dtype"])
        dim = info_per_bit[v]["dim"]
        dim_name = f"bit{dtype}"
        dsb[v] = xr.DataArray(
            info_per_bit[v]["bitinfo"],
            dims=[dim_name],
            coords={dim_name: get_bit_coords(dtype), "dim": dim},
            name=v,
            attrs={
                "long_name": f"{v} bitwise information",
                "units": 1,
            },
        ).astype("float64")
    # add metadata
    dsb.attrs = {
        "xbitinfo_description": "bitinformation calculated by xbitinfo.get_bitinformation wrapping bitinformation.jl",
        "python_repository": "https://github.com/observingClouds/xbitinfo",
        "julia_repository": "https://github.com/milankl/BitInformation.jl",
        "reference_paper": "http://www.nature.com/articles/s43588-021-00156-2",
        "xbitinfo_version": xb.__version__,
        "BitInformation.jl_version": get_julia_package_version("BitInformation"),
    }
    for c in dsb.coords:
        if "bit" in c:
            dsb.coords[c].attrs = {
                "description": "name of the bits: '±' refers to the sign bit, 'e' to the exponents bits and 'm' to the mantissa bits."
            }
    dsb.coords["dim"].attrs = {
        "description": "dimension of the source dataset along which the bitwise information has been analysed."
    }
    return dsb


def _check_bitinfo_kwargs(implementation=None, axis=None, dim=None, kwargs=None):
    if kwargs is None:
        kwargs = {}
    # check keywords
    if implementation == "julia" and not julia_installed:
        raise ImportError('Please install julia or use implementation="python".')
    if axis is not None and dim is not None:
        raise ValueError("Please provide either `axis` or `dim` but not both.")
    if axis:
        if not isinstance(axis, int):
            raise ValueError(f"Please provide `axis` as `int`, found {type(axis)}.")
    if dim:
        if not isinstance(dim, str) and not isinstance(dim, list):
            raise ValueError(
                f"Please provide `dim` as `str` or `list`, found {type(dim)}."
            )
    if "mask" in kwargs:
        raise ValueError(
            "`xbitinfo` does not wrap the mask argument. Mask your xr.Dataset with NaNs instead."
        )
    return


def get_bitinformation(
    ds,
    dim=None,
    axis=None,
    label=None,
    overwrite=False,
    implementation="julia",
    **kwargs,
):
    """Wrap `BitInformation.jl.bitinformation() <https://github.com/milankl/BitInformation.jl/blob/main/src/mutual_information.jl>`__.

    Parameters
    ----------
    ds : :py:class:`xarray.Dataset`
      Input dataset to analyse
    dim : str or list
      Dimension over which to apply mean. Only one of the ``dim`` and ``axis`` arguments can be supplied.
      If no ``dim`` or ``axis`` is given (default), the bitinformation is retrieved along all dimensions.
    axis : int
      Axis over which to apply mean. Only one of the ``dim`` and ``axis`` arguments can be supplied.
      If no ``dim`` or ``axis`` is given (default), the bitinformation is retrieved along all dimensions.
    label : str
      Label of the json to serialize bitinfo. When string, serialize results to disk into file ``{{label}}.json`` to be reused later. Defaults to ``None``.
    overwrite : bool
      If ``False``, try using serialized bitinfo based on label; if true or label does not exist, run bitinformation
    implementation : str
      Bitinformation algorithm implementation. Valid options are
        - julia, the original implementation of julia in julia by Milan Kloewer
        - python, a copy of the core functionality of julia in python
    kwargs
      to be passed to bitinformation:

        - masked_value: defaults to ``NaN`` (different to ``julia`` defaulting to ``"nothing"``), set ``None`` disable masking
        - mask: use ``masked_value`` instead
        - set_zero_insignificant (``bool``): defaults to ``True`` (julia implementation) or ``False`` (python implementation)
        - confidence (``float``): defaults to ``0.99``


    Returns
    -------
    info_per_bit : :py:class:`xarray.Dataset`
      Information content per ``bit`` and ``variable`` (and ``dim``)

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> xb.get_bitinformation(ds, dim="lon")  # doctest: +ELLIPSIS
    <xarray.Dataset> Size: 1kB
    Dimensions:     (bitfloat64: 64)
    Coordinates:
      * bitfloat64  (bitfloat64) <U3 768B '±' 'e1' 'e2' 'e3' ... 'm50' 'm51' 'm52'
        dim         <U3 12B 'lon'
    Data variables:
        air         (bitfloat64) float64 512B 0.0 0.0 0.0 ... 0.002848 0.0 0.0005048
    Attributes:
        xbitinfo_description:       bitinformation calculated by xbitinfo.get_bit...
        python_repository:          https://github.com/observingClouds/xbitinfo
        julia_repository:           https://github.com/milankl/BitInformation.jl
        reference_paper:            http://www.nature.com/articles/s43588-021-001...
        xbitinfo_version:           ...
        BitInformation.jl_version:  ...
    >>> xb.get_bitinformation(ds)
    <xarray.Dataset> Size: 2kB
    Dimensions:     (bitfloat64: 64, dim: 3)
    Coordinates:
      * bitfloat64  (bitfloat64) <U3 768B '±' 'e1' 'e2' 'e3' ... 'm50' 'm51' 'm52'
      * dim         (dim) <U4 48B 'lat' 'lon' 'time'
    Data variables:
        air         (dim, bitfloat64) float64 2kB 0.0 0.0 0.0 ... 0.0 0.0004506
    Attributes:
        xbitinfo_description:       bitinformation calculated by xbitinfo.get_bit...
        python_repository:          https://github.com/observingClouds/xbitinfo
        julia_repository:           https://github.com/milankl/BitInformation.jl
        reference_paper:            http://www.nature.com/articles/s43588-021-001...
        xbitinfo_version:           ...
        BitInformation.jl_version:  ...
    """
    if overwrite is False and label is not None:
        try:
            info_per_bit = load_bitinformation(label)
        except FileNotFoundError:
            logging.info(
                f"No bitinformation could be found for {label}. Please set `overwrite=True` for recalculation..."
            )
        else:
            return info_per_bit
    else:
        _check_bitinfo_kwargs(implementation, axis, dim, kwargs)

    return _get_bitinformation(
        ds,
        dim=dim,
        axis=axis,
        label=label,
        overwrite=overwrite,
        implementation=implementation,
        **kwargs,
    )


def _get_bitinformation(
    ds,
    dim=None,
    axis=None,
    label=None,
    overwrite=False,
    implementation="julia",
    **kwargs,
):
    if dim is None and axis is None:
        # gather bitinformation on all axis
        return _get_bitinformation_along_dims(
            ds,
            dim=dim,
            label=label,
            overwrite=overwrite,
            implementation=implementation,
            **kwargs,
        )
    if isinstance(dim, list) and axis is None:
        # gather bitinformation on dims specified
        return _get_bitinformation_along_dims(
            ds,
            dim=dim,
            label=label,
            overwrite=overwrite,
            implementation=implementation,
            **kwargs,
        )
    else:
        # gather bitinformation along one axis
        info_per_bit = _get_bitinformation_along_axis(
            ds, implementation, axis, dim, **kwargs
        )

        if label is not None:
            out_fn = label + ".json"
            if not os.path.exists(out_fn) or overwrite:
                save_bitinformation(info_per_bit, out_fn)

        info_per_bit = dict_to_dataset(info_per_bit)

    for var in info_per_bit.data_vars:  # keep attrs from input with source_ prefix
        for a in ds[var].attrs.keys():
            info_per_bit[var].attrs["source_" + a] = ds[var].attrs[a]
    return info_per_bit


def _quantized_variable_is_scaled(ds: xr.DataArray, var: str) -> bool:
    has_scale_or_offset = any(
        ["add_offset" in ds[var].encoding, "scale_factor" in ds[var].encoding]
    )

    if not has_scale_or_offset:
        return False

    loaded_dtype = ds[var].dtype
    storage_dtype = ds[var].encoding.get("dtype", None)
    assert (
        storage_dtype is not None
    ), f"Variable {var} is likely quantized, but does not have a storage dtype"

    if loaded_dtype == storage_dtype:
        return False

    return True


def _jl_get_bitinformation(ds, var, axis, dim, kwargs={}):
    X = ds[var].values
    Main.X = X
    if axis is not None:
        # in julia convention axis + 1
        axis_jl = axis + 1
        dim = ds[var].dims[axis]
    if isinstance(dim, str):
        try:
            # in julia convention axis + 1
            axis_jl = ds[var].get_axis_num(dim) + 1
        except ValueError:
            logging.info(f"Variable {var} does not have dimension {dim}. Skipping.")
            return
    assert isinstance(axis_jl, int)
    Main.dim = axis_jl
    kwargs_str = _get_bitinformation_kwargs_handler(ds[var], kwargs)
    logging.debug(f"get_bitinformation(X, dim={dim}, {kwargs_str})")
    info_per_bit = {}
    info_per_bit["bitinfo"] = jl.eval(
        f"get_bitinformation(X, dim={axis_jl}, {kwargs_str})"
    )
    info_per_bit["dim"] = dim
    info_per_bit["axis"] = axis_jl - 1
    info_per_bit["dtype"] = str(ds[var].dtype)
    return info_per_bit


def _py_get_bitinformation(ds, var, axis, dim, kwargs={}):
    if "set_zero_insignificant" in kwargs.keys():
        if kwargs["set_zero_insignificant"]:
            raise NotImplementedError(
                "set_zero_insignificant is not implemented in the python implementation"
            )
    else:
        assert (
            kwargs == {}
        ), "This implementation only supports the plain bitinfo implementation"
    itemsize = ds[var].dtype.itemsize
    astype = f"u{itemsize}"
    X = da.array(ds[var])

    # signed exponent conversion only for floats
    if X.dtype in (np.float16, np.float32, np.float64):
        X = pb.signed_exponent(X)

    X = X.astype(astype)
    if axis is not None:
        dim = ds[var].dims[axis]
    if isinstance(dim, str):
        try:
            axis = ds[var].get_axis_num(dim)
        except ValueError:
            logging.info(f"Variable {var} does not have dimension {dim}. Skipping.")
            return
    info_per_bit = {}
    logging.info("Calling python implementation now")
    info_per_bit["bitinfo"] = pb.bitinformation(X, axis=axis).compute()
    info_per_bit["dim"] = dim
    info_per_bit["axis"] = axis
    info_per_bit["dtype"] = str(ds[var].dtype)
    return info_per_bit


def _get_bitinformation_along_dims(
    ds,
    dim=None,
    label=None,
    overwrite=False,
    implementation="julia",
    **kwargs,
):
    """Helper function for :py:func:`xbitinfo.xbitinfo.get_bitinformation` to handle multi-dimensional analysis for each dim specified.

    Simple wrapper around :py:func:`xbitinfo.xbitinfo.get_bitinformation`, which calls :py:func:`xbitinfo.xbitinfo.get_bitinformation`
    for each dimension found in the provided :py:func:`xarray.Dataset`. The retrieved bitinformation
    is gathered in a joint :py:func:`xarray.Dataset` and is returned.
    """
    info_per_bit_per_dim = {}
    if dim is None:
        dim = ds.dims
    for d in dim:
        logging.info(f"Get bitinformation along dimension {d}")
        if label is not None:
            label = "_".join([label, d])
        info_per_bit_per_dim[d] = _get_bitinformation(
            ds,
            dim=d,
            axis=None,
            label=label,
            overwrite=overwrite,
            implementation=implementation,
            **kwargs,
        ).expand_dims("dim", axis=0)
    info_per_bit = xr.merge(
        info_per_bit_per_dim.values(), join="outer", compat="no_conflicts"
    ).squeeze()
    return info_per_bit


def _get_bitinformation_along_axis(ds, implementation, axis, dim, **kwargs):
    """
    Helper function for :py:func:`xbitinfo.xbitinfo.get_bitinformation` to handle analysis along one axis.
    """
    info_per_bit = {}
    pbar = tqdm(ds.data_vars)
    for var in pbar:
        pbar.set_description(f"Processing var: {var} for dim: {dim}")
        if _quantized_variable_is_scaled(ds, var):
            loaded_dtype = ds[var].dtype
            quantized_storage_dtype = ds[var].encoding["dtype"]
            warnings.warn(
                f"Variable {var} is quantized as {quantized_storage_dtype}, but loaded as {loaded_dtype}. Consider reopening using `mask_and_scale=False` to get sensible results",
                category=UserWarning,
            )
        if implementation == "julia":
            info_per_bit_var = _jl_get_bitinformation(ds, var, axis, dim, kwargs)
            if info_per_bit_var is None:
                continue
            else:
                info_per_bit[var] = info_per_bit_var
        elif implementation == "python":
            info_per_bit_var = _py_get_bitinformation(ds, var, axis, dim, kwargs)
            if info_per_bit_var is None:
                continue
            else:
                info_per_bit[var] = info_per_bit_var
        else:
            raise ValueError(
                f"Implementation of bitinformation algorithm {implementation} is unknown. Please choose a different one."
            )

    return info_per_bit


def _get_bitinformation_kwargs_handler(da, kwargs):
    """Helper function to preprocess kwargs args of :py:func:`xbitinfo.xbitinfo.get_bitinformation`."""
    kwargs_var = kwargs.copy()
    if "masked_value" not in kwargs_var:
        if da.dtype.kind == "i" or da.dtype.kind == "u":
            logging.warning(
                "No masked value given for integer type variable. Assuming no mask to apply."
            )
            kwargs_var["masked_value"] = "nothing"
        elif da.dtype.kind == "f":
            kwargs_var["masked_value"] = f"convert({str(da.dtype).capitalize()},NaN)"
        else:
            raise ValueError(f"Dtype kind ({da.dtype.kind}) not supported.")
    elif kwargs_var["masked_value"] is None:
        kwargs_var["masked_value"] = "nothing"
    if "set_zero_insignificant" not in kwargs_var:
        kwargs_var["set_zero_insignificant"] = True
    kwargs_str = ", ".join([f"{k}={v}" for k, v in kwargs_var.items()])
    # convert python to julia bool
    kwargs_str = kwargs_str.replace("True", "true").replace("False", "false")
    return kwargs_str


def load_bitinformation(label):
    """Load bitinformation from JSON file"""
    label_file = label + ".json"
    if os.path.exists(label_file):
        with open(label_file) as f:
            logging.debug(f"Load bitinformation from {label+'.json'}")
            info_per_bit = json.load(f)
        return dict_to_dataset(info_per_bit)
    else:
        raise FileNotFoundError(f"No bitinformation could be found at {label+'.json'}")


def get_cdf_without_artificial_information(
    info_per_bit, bitdim, threshold, tolerance, bit_vars
):
    """
    Calculate a Cumulative Distribution Function (CDF) with artificial information removal.

    This function calculates a modified CDF for a given set of bit information and variable dimensions,
    removing artificial information while preserving the desired threshold of information content.

    1.)The function's aim is to return the cdf in a way that artificial information gets removed.
    2.)This function calculates the CDF using the provided information content per bit dataset.
    3.)It then computes the gradient of the CDF values to identify points where the gradient becomes close to the given tolerance,
    indicating a drop in information.
    4.)Simultaneously, it keeps track of the minimum cumulative sum of information content which is threshold here, which signifies atleast
    this much fraction of total information needs to be passed.
    5.)So the bit where the intersection of the gradient reaching the tolerance and the cumulative sum exceeding the threshold. All bits beyond this
    index are assumed to contain artificial information and are set to zero in the resulting CDF.


    Parameters:
    -----------
    info_per_bit : :py:class: 'xarray.Dataset'
        Information content of each bit. This is the output from :py:func:`xbitinfo.xbitinfo.get_bitinformation`.
    bitdim : str
        The dimension representing the bit information.
    threshold : float
        Minimum cumulative sum of information content before artificial information filter is applied.
    tolerance : float
        The tolerance is the value below which gradient starts becoming constant
    bit_vars : list
        List of variable names of the dataset.

    Returns:
    --------
    xarray.Dataset
        A modified CDF dataset with artificial information removed.

    Example:
    --------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info = xb.get_bitinformation(ds)
    >>> get_keepbits(
    ...     info,
    ...     inflevel=[0.99],
    ...     information_filter="Gradient",
    ...     **{"threshold": 0.7, "tolerance": 0.001}
    ... )
    <xarray.Dataset> Size: 80B
    Dimensions:   (dim: 3, inflevel: 1)
    Coordinates:
      * dim       (dim) <U4 48B 'lat' 'lon' 'time'
      * inflevel  (inflevel) float64 8B 0.99
    Data variables:
        air       (dim, inflevel) int64 24B 5 7 6
    ...
    """

    # Extract coordinates from the 'info_per_bit' dataset.
    coordinates = info_per_bit.coords
    # Extract the 'dim' values from the coordinates and store them in 'coordinates_array'.
    coordinates_array = coordinates["dim"].values
    # Initialize a flag to identify if 'coordinates_array' is a scalar value.
    flag_scalar_value = False
    # Check if 'coordinates_array' is a scalar (has zero dimensions).
    if coordinates_array.ndim == 0:
        # If it's a scalar, extract the scalar value and set the flag to True.
        value = coordinates_array.item()
        flag_scalar_value = True
        # Convert the scalar value into a 1D numpy array so that we can iterate over it for determining dimensions.
        coordinates_array = np.array([value])

    cdf = _cdf_from_info_per_bit(info_per_bit, bitdim)
    for var_name in bit_vars:
        for dimension in coordinates_array:
            if flag_scalar_value:
                # If it's a scalar, extract the information array directly.
                infoArray = info_per_bit[var_name]
            else:
                # If it's not a scalar, select the information array using the specified dimension.
                infoArray = info_per_bit[var_name].sel(dim=dimension)

            # total sum of information along a dimension
            infSum = sum(infoArray).item()

            data_type = np.dtype(bitdim.replace("bit", ""))
            _, n_sign, n_exponent, _ = bit_partitioning(data_type)
            sign_and_exponent = n_sign + n_exponent

            # sum of sign and exponent bits
            SignExpSum = sum(infoArray[:sign_and_exponent]).item()
            if flag_scalar_value:
                cdf_array = cdf[var_name]
            else:
                cdf_array = cdf[var_name].sel(dim=dimension)

            gradient_array = np.diff(cdf_array.values)
            # Initialize 'CurrentBit_Sum' with the value of 'SignExpSum'.
            CurrentBit_Sum = SignExpSum
            for i in range(sign_and_exponent, len(gradient_array) - 1):
                # Update 'CurrentBit_Sum' by adding the information content of the current bit.
                CurrentBit_Sum = CurrentBit_Sum + infoArray[i].item()
                if (
                    gradient_array[i]
                ) < tolerance and CurrentBit_Sum >= threshold * infSum:
                    infbits = i
                    break

            for i in range(0, infbits + 1):
                # Normalize CDF values for elements up to 'infbits'.
                cdf_array[i] = cdf_array[i] / cdf_array[infbits]

            cdf_array[(infbits + 1) :] = 1
    return cdf


def save_bitinformation(info_per_bit, out_fn, overwrite=False):
    """Save bitinformation to JSON file"""
    with open(out_fn, "w") as f:
        logging.debug(f"Save bitinformation to {out_fn}")
        json.dump(info_per_bit, f, cls=JsonCustomEncoder)
    return


def get_keepbits(info_per_bit, inflevel=0.99, information_filter=None, **kwargs):
    """Get the number of mantissa bits to keep. To be used in :py:func:`xbitinfo.bitround.xr_bitround` and :py:func:`xbitinfo.bitround.jl_bitround`.

    Parameters
    ----------
    info_per_bit : :py:class:`xarray.Dataset`
      Information content of each bit. This is the output from :py:func:`xbitinfo.xbitinfo.get_bitinformation`.
    inflevel : float or list
      Level of information that shall be preserved.

    Kwargs
        threshold(` `float ``) : defaults to ``0.7``
            Minimum cumulative sum of information content before artificial information filter is applied.
        tolerance(` `float ``) : defaults to ``0.001``
            The tolerance is the value below which gradient starts becoming constant


    Returns
    -------
    keepbits : dict
      Number of mantissa bits to keep per variable

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info_per_bit = xb.get_bitinformation(ds, dim="lon")
    >>> xb.get_keepbits(info_per_bit)
    <xarray.Dataset> Size: 28B
    Dimensions:   (inflevel: 1)
    Coordinates:
      * inflevel  (inflevel) float64 8B 0.99
        dim       <U3 12B 'lon'
    Data variables:
        air       (inflevel) int64 8B 7
    ...
    >>> xb.get_keepbits(info_per_bit, inflevel=0.99999999)
    <xarray.Dataset> Size: 28B
    Dimensions:   (inflevel: 1)
    Coordinates:
      * inflevel  (inflevel) float64 8B 1.0
        dim       <U3 12B 'lon'
    Data variables:
        air       (inflevel) int64 8B 7
    ...
    >>> xb.get_keepbits(info_per_bit, inflevel=1.0)
    <xarray.Dataset> Size: 28B
    Dimensions:   (inflevel: 1)
    Coordinates:
      * inflevel  (inflevel) float64 8B 1.0
        dim       <U3 12B 'lon'
    Data variables:
        air       (inflevel) int64 8B 52
    ...
    >>> info_per_bit = xb.get_bitinformation(ds)
    >>> xb.get_keepbits(info_per_bit)
    <xarray.Dataset> Size: 80B
    Dimensions:   (dim: 3, inflevel: 1)
    Coordinates:
      * dim       (dim) <U4 48B 'lat' 'lon' 'time'
      * inflevel  (inflevel) float64 8B 0.99
    Data variables:
        air       (dim, inflevel) int64 24B 5 7 6
    ...
    """
    if not isinstance(inflevel, list):
        inflevel = [inflevel]
    keepmantissabits = []
    inflevel = xr.DataArray(inflevel, dims="inflevel", coords={"inflevel": inflevel})
    if (inflevel < 0).any() or (inflevel > 1.0).any():
        raise ValueError("Please provide `inflevel` from interval [0.,1.]")
    for bitdim in [
        "bitfloat16",
        "bitfloat32",
        "bitfloat64",
        "bitint16",
        "bitint32",
        "bitint64",
        "bituint16",
        "bituint32",
        "bituint64",
    ]:
        # get only variables of bitdim
        bit_vars = [v for v in info_per_bit.data_vars if bitdim in info_per_bit[v].dims]
        if bit_vars != []:
            if information_filter == "Gradient":
                cdf = get_cdf_without_artificial_information(
                    info_per_bit[bit_vars],
                    bitdim,
                    kwargs["threshold"],
                    kwargs["tolerance"],
                    bit_vars,
                )
            else:
                cdf = _cdf_from_info_per_bit(info_per_bit[bit_vars], bitdim)
            data_type = np.dtype(bitdim.replace("bit", ""))
            n_bits, _, _, n_mant = bit_partitioning(data_type)
            bitdim_non_mantissa_bits = n_bits - n_mant

            keepmantissabits_bitdim = (
                (cdf > inflevel).argmax(bitdim) + 1 - bitdim_non_mantissa_bits
            )
            # keep all mantissa bits for 100% information
            if 1.0 in inflevel:
                bitdim_all_mantissa_bits = n_bits - bitdim_non_mantissa_bits
                keepall = xr.ones_like(keepmantissabits_bitdim.sel(inflevel=1.0)) * (
                    bitdim_all_mantissa_bits
                )
                keepmantissabits_bitdim = xr.concat(
                    [keepmantissabits_bitdim.drop_sel(inflevel=1.0), keepall],
                    "inflevel",
                )
            keepmantissabits.append(keepmantissabits_bitdim)
    keepmantissabits = xr.merge(keepmantissabits, join="outer", compat="no_conflicts")
    if inflevel.inflevel.size > 1:  # restore original ordering
        keepmantissabits = keepmantissabits.sel(inflevel=inflevel.inflevel)
    return keepmantissabits


def _cdf_from_info_per_bit(info_per_bit, bitdim):
    """Convert info_per_bit to cumulative distribution function on dimension bitdim."""
    # set below rounding error from last digit to zero
    info_per_bit_cleaned = info_per_bit.where(
        info_per_bit > info_per_bit.isel({bitdim: slice(-4, None)}).max(bitdim) * 1.5
    )
    # make cumulative distribution function
    cdf = info_per_bit_cleaned.cumsum(bitdim) / info_per_bit_cleaned.cumsum(
        bitdim
    ).isel({bitdim: -1})
    return cdf


def _jl_bitround(X, keepbits):
    """Wrap `BitInformation.jl.round <https://github.com/milankl/BitInformation.jl/blob/main/src/round_nearest.jl>`__. Used in :py:func:`xbitinfo.bitround.jl_bitround`."""
    if not julia_installed:
        raise ImportError("Please install julia or use xr_bitround")
    Main.X = X
    Main.keepbits = keepbits
    return jl.eval("round!(X, keepbits)")


def get_prefect_flow(paths=[]):
    """
    Create `prefect.Flow <https://docs.prefect.io/core/concepts/flows.html#overview>`__ for paths to be:

    1. Analyse bitwise real information content with :py:func:`xbitinfo.xbitinfo.get_bitinformation`
    2. Retrieve keepbits with :py:func:`xbitinfo.xbitinfo.get_keepbits`
    3. Apply bitrounding with :py:func:`xbitinfo.bitround.xr_bitround`
    4. Save as compressed netcdf with :py:class:`xbitinfo.save_compressed.ToCompressed_Netcdf`

    Many parameters can be changed when running the flow ``flow.run(parameters=dict(chunk="auto"))``:
    - paths: list of paths
        Paths to be bitrounded
    - analyse_paths: str or int
        Which paths to be passed to :py:func:`xbitinfo.xbitinfo.get_bitinformation`. Choose from ``["first_last", "all", int]``, where int is interpreted as stride, i.e. paths[::stride]. Defaults to ``"first"``.
    - enforce_dtype : str or None
        Enforce dtype for all variables. Currently, :py:func:`xbitinfo.xbitinfo.get_bitinformation` fails for different dtypes in variables. Do nothing if ``None``. Defaults to ``None``.
    - label : see :py:func:`xbitinfo.xbitinfo.get_bitinformation`
    - dim/axis : see :py:func:`xbitinfo.xbitinfo.get_bitinformation`
    - inflevel : see :py:func:`xbitinfo.xbitinfo.get_keepbits`
    - non_negative_keepbits : bool
        Set negative keepbits from :py:func:`xbitinfo.xbitinfo.get_keepbits` to ``0``. Required when using :py:func:`xbitinfo.bitround.xr_bitround`. Defaults to True.
    - chunks : see :py:meth:`xarray.open_mfdataset`. Note that with ``chunks=None``, ``dask`` is not used for I/O and the flow is still parallelized when using ``DaskExecutor``.
    - bitround_in_julia : bool
        Use :py:func:`xbitinfo.bitround.jl_bitround` instead of :py:func:`xbitinfo.bitround.xr_bitround`. Both should yield identical results. Defaults to ``False``.
    - overwrite : bool
        Whether to overwrite bitrounded netcdf files. ``False`` (default) skips existing files.
    - complevel : see to_compressed_netcdf, defaults to ``7``.
    - rename : list
        Replace mapping for paths towards new_path of bitrounded file, i.e. ``replace=[".nc", "_bitrounded_compressed.nc"]``

    Parameters
    ------
    paths : list
      List of paths of files to be processed by :py:func:`xbitinfo.xbitinfo.get_bitinformation`, :py:func:`xbitinfo.xbitinfo.get_keepbits`, :py:func:`xbitinfo.bitround.xr_bitround` and ``to_compressed_netcdf``.

    Returns
    -------
    prefect.Flow
      See https://docs.prefect.io/core/concepts/flows.html#overview

    Example
    -------
    Imagine n files of identical structure, i.e. 1-year per file climate model output:

    >>> ds = xr.tutorial.load_dataset("rasm")
    >>> year, datasets = zip(*ds.groupby("time.year"))
    >>> paths = [f"{y}.nc" for y in year]
    >>> xr.save_mfdataset(datasets, paths)

    Create prefect.Flow and run sequentially

    >>> flow = xb.get_prefect_flow(paths=paths)
    >>> import prefect
    >>> from prefect import get_run_logger
    >>> if __name__ == "__main__":
    ...     logger = get_run_logger()
    ...     logger.setLevel("ERROR")
    ...     st = flow.run()
    ...

    Inspect flow state

    >>> # flow.visualize(st)  # requires graphviz

    Run in parallel with dask:

    >>> import os  # https://docs.xarray.dev/en/stable/user-guide/dask.html
    >>> os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
    >>> from prefect_dask.task_runners import DaskTaskRunner
    >>> from dask.distributed import Client
    >>> if __name__ == "__main__":
    ...     client = Client(n_workers=2, threads_per_worker=1, processes=True)
    ...     executor = DaskTaskRunner(
    ...         address=client.scheduler.address
    ...     )  # take your own client
    ...     flow.run(executor=executor, parameters=dict(overwrite=True))
    ...

    Modify parameters of a flow:

    >>> if __name__ == "__main__":
    ...     flow.run(parameters=dict(inflevel=0.9999, overwrite=True))
    ...

    See also
    --------
    `ETL Pipelines with Prefect <https://examples.dask.org/applications/prefect-etl.html/>`__ and
    `Run a flow <https://docs.prefect.io/core/getting_started/basic-core-flow.html>`__
    """

    from .bitround import jl_bitround, xr_bitround

    @task
    def get_bitinformation_keepbits(
        paths,
        analyse_paths="first",
        label=None,
        inflevel=0.99,
        enforce_dtype=None,
        non_negative_keepbits=True,
        **get_bitinformation_kwargs,
    ):
        # take subset only for analysis in bitinformation
        if analyse_paths == "first_last":
            p = [paths[0], paths[-1]]
        elif analyse_paths == "all":
            p = paths
        elif analyse_paths == "first":
            p = paths[0]
        elif isinstance(analyse_paths, int):  # interpret as stride
            p = paths[::analyse_paths]
        else:
            raise ValueError(
                "Please provide analyse_paths as int interpreted as stride or from ['first_last','all','first','last']."
            )
        ds = xr.open_mfdataset(p)
        if enforce_dtype:
            ds = ds.astype(enforce_dtype)
        info_per_bit = get_bitinformation(ds, label=label, **get_bitinformation_kwargs)
        keepbits = get_keepbits(info_per_bit, inflevel=inflevel)
        if non_negative_keepbits:
            keepbits = {v: max(0, k) for v, k in keepbits.items()}  # ensure no negative
        return keepbits

    @task
    def bitround_and_save(
        path,
        keepbits,
        chunks=None,
        complevel=4,
        rename=[".nc", "_bitrounded_compressed.nc"],
        overwrite=False,
        enforce_dtype=None,
        bitround_in_julia=False,
    ):
        new_path = path.replace(rename[0], rename[1])
        if not overwrite:
            if os.path.exists(new_path):
                try:
                    ds_new = xr.open_dataset(new_path, chunks=chunks)
                    ds = xr.open_dataset(path, chunks=chunks)
                    if ds.nbytes == ds_new.nbytes:
                        raise ValueError(f"{new_path} already exists.")
                except Exception as e:
                    print(
                        f"{type(e)} when xr.open_dataset({new_path}), therefore delete and recalculate."
                    )
                    os.remove(new_path)

        ds = xr.open_dataset(path, chunks=chunks)
        if enforce_dtype:
            ds = ds.astype(enforce_dtype)

        bitround_func = jl_bitround if bitround_in_julia else xr_bitround
        ds_bitround = bitround_func(ds, keepbits)
        ds_bitround.to_compressed_netcdf(new_path, complevel=complevel)

    @flow
    def xbitinfo_pipeline(
        paths,
        analyse_paths="first",
        dim=None,
        axis=0,
        inflevel=0.99,
        label=None,
        rename=[".nc", "_bitrounded_compressed.nc"],
        overwrite=False,
        bitround_in_julia=False,
        complevel=7,
        chunks=None,
        enforce_dtype=None,
        non_negative_keepbits=True,
    ):
        if not paths:
            raise ValueError("Please provide paths of files to bitround, found [].")

        keepbits = get_bitinformation_keepbits(
            paths,
            analyse_paths=analyse_paths,
            dim=dim,
            axis=axis,
            inflevel=inflevel,
            label=label,
            enforce_dtype=enforce_dtype,
            non_negative_keepbits=non_negative_keepbits,
        )

        # Parallel map
        bitround_and_save.map(
            paths,
            keepbits=unmapped(keepbits),
            rename=unmapped(rename),
            chunks=unmapped(chunks),
            complevel=unmapped(complevel),
            overwrite=unmapped(overwrite),
            enforce_dtype=unmapped(enforce_dtype),
            bitround_in_julia=unmapped(bitround_in_julia),
        )


class JsonCustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (np.ndarray, np.number)):
            return obj.tolist()
        elif isinstance(obj, (complex, np.complex)):
            return [obj.real, obj.imag]
        elif isinstance(obj, set):
            return list(obj)
        elif isinstance(obj, bytes):  # pragma: py3
            return obj.decode()
        return json.JSONEncoder.default(self, obj)


def get_julia_package_version(package):
    """Get version information of julia package"""
    if julia_installed:
        version = jl.eval(
            f'Pkg.TOML.parsefile(joinpath(pkgdir({package}), "Project.toml"))["version"]'
        )
    else:
        version = "implementation='python'"
    return version



import numpy as np
import pytest
import xarray as xr
from dask import is_dask_collection
from xarray.testing import assert_allclose, assert_equal

import xbitinfo as xb
from xbitinfo import bitround as bi

from . import requires_julia


@pytest.mark.parametrize("dtype", ["float16", "float32", "float64"])
@pytest.mark.parametrize("implementation", ["xarray", "julia"])
@pytest.mark.parametrize("input_type", ["Dataset", "DataArray"])
@pytest.mark.parametrize("keepbits", ["dict", "int"])
def test_xr_bitround(air_temperature, dtype, input_type, implementation, keepbits):
    """Test xr_bitround to different keepbits of type dict or int."""
    ds = air_temperature.astype(dtype)
    i = 6
    if keepbits == "dict":
        keepbits = dict.fromkeys(ds.data_vars, i)
    elif keepbits == "int":
        keepbits = i
    if input_type == "DataArray":
        v = list(ds.data_vars)[0]
        ds = ds[v]

    bitround = xb.xr_bitround if implementation == "xarray" else xb.jl_bitround
    ds_bitrounded = bitround(ds, keepbits)

    def check(da, da_bitrounded):
        # check close
        assert_allclose(da, da_bitrounded, atol=0.01, rtol=0.01)
        # attrs set
        assert da_bitrounded.attrs["_QuantizeBitRoundNumberOfSignificantDigits"] == i
        # different after bitrounding
        diff = (da - da_bitrounded).compute()
        assert (diff != 0).any()

    if input_type == "DataArray":
        check(ds, ds_bitrounded)
    else:
        for v in ds.data_vars:
            check(ds[v], ds_bitrounded[v])


@pytest.mark.parametrize(
    "implementation,dask",
    [("xarray", True), ("xarray", False), ("julia", False)],
)
def test_bitround_dask(air_temperature, implementation, dask):
    """Test xr_bitround and jl_bitround keeps dask and successfully computes."""
    ds = air_temperature
    i = 15
    keepbits = i
    if dask:
        ds = ds.chunk("auto")

    bitround = xb.xr_bitround if implementation == "xarray" else xb.jl_bitround
    ds_bitrounded = bitround(ds, keepbits)
    assert is_dask_collection(ds_bitrounded) == dask
    if dask:
        assert ds_bitrounded.compute()


@requires_julia
@pytest.mark.parametrize(
    "dtype,keepbits",
    [("float16", range(1, 9)), ("float32", range(1, 23)), ("float64", range(1, 52))],
)
def test_bitround_xarray_julia_equal(air_temperature, dtype, keepbits):
    """Test jl_bitround and xr_bitround yield identical results."""
    ds = air_temperature.astype(dtype)
    for keep in keepbits:
        ds_xr_bitrounded = xb.xr_bitround(ds, keep)
        ds_jl_bitrounded = xb.jl_bitround(ds, keep)
        assert_equal(ds_jl_bitrounded, ds_xr_bitrounded)


def test_bitround_along_dim(air_temperature):
    # test for inflevels
    ds = air_temperature
    info_per_bit = xb.get_bitinformation(ds, dim="lon")
    ds_bitrounded_along_lon = bi.bitround_along_dim(
        ds, info_per_bit, dim="lon", inflevels=[1.0, 0.9999, 0.99, 0.975]
    )

    assert ds_bitrounded_along_lon.air.dtype == "float64"
    assert ds_bitrounded_along_lon.lon.size == ds.lon.size
    assert ds_bitrounded_along_lon.lat.size == ds.lat.size
    assert ds_bitrounded_along_lon.time.size == ds.time.size
    assert ds.air.values.dtype == ds_bitrounded_along_lon.air.values.dtype

    assert (ds - ds_bitrounded_along_lon).air.mean() < 0.01

    # test for keepbits
    ds_bitrounded_along_lon = bi.bitround_along_dim(
        ds, info_per_bit, dim="lon", inflevels=None, keepbits=2
    )

    assert ds_bitrounded_along_lon.air.dtype == "float64"
    assert ds_bitrounded_along_lon.lon.size == ds.lon.size
    assert ds_bitrounded_along_lon.lat.size == ds.lat.size
    assert ds_bitrounded_along_lon.time.size == ds.time.size
    assert ds.air.values.dtype == ds_bitrounded_along_lon.air.values.dtype

    assert (ds - ds_bitrounded_along_lon).air.mean() < 0.01

    # Test error when both keepbits and inflevels are provided
    with pytest.raises(ValueError):
        bi.bitround_along_dim(
            ds,
            info_per_bit,
            dim="lat",
            keepbits=2,
            inflevels=[1.0, 0.9999, 0.99, 0.975],
        )

    # Test error when neither keepbits nor inflevels are provided
    with pytest.raises(ValueError):
        bi.bitround_along_dim(ds, info_per_bit, dim="lat", inflevels=None)



import os
import shutil

import pytest
import xarray as xr
import zarr
from packaging.version import Version

zarr_version = Version(zarr.__version__)

if zarr_version >= Version("3.0.0"):
    from zarr.codecs import BloscCodec, BloscShuffle

    compressor_key = "compressors"
else:
    from zarr.codecs import Blosc as BloscCodec

    compressor_key = "compressor"

    class BloscShuffle:
        bitshuffle = zarr.codecs.blosc.BITSHUFFLE
        shuffle = zarr.codecs.blosc.SHUFFLE
        noshuffle = zarr.codecs.blosc.NOSHUFFLE


import xbitinfo as xb


@pytest.mark.parametrize("for_cdo", [True, False])
def test_get_compress_encoding_for_cdo(rasm, for_cdo):
    ds = rasm
    encoding = xb.get_compress_encoding_nc(ds, for_cdo=for_cdo)
    v = list(ds.data_vars)[0]
    time_axis = ds[v].get_axis_num("time")
    if for_cdo:
        assert encoding[v]["chunksizes"][time_axis] == 1
    else:
        assert encoding[v]["chunksizes"][time_axis] > 1


@pytest.mark.parametrize("dask", [True, False])
def test_to_compressed_netcdf(rasm, dask):
    """Test to_compressed_netcdf reduces size on disk."""
    ds = rasm
    if dask:
        ds = ds.chunk("auto")
    label = "file"
    # save
    ds.to_netcdf(f"./tmp_testdir/{label}.nc")
    ds.to_compressed_netcdf(f"./tmp_testdir/{label}_compressed.nc")
    # check size reduction
    ori_size = os.path.getsize(f"./tmp_testdir/{label}.nc")
    compressed_size = os.path.getsize(f"./tmp_testdir/{label}_compressed.nc")
    assert compressed_size < ori_size


def test_to_compressed_netcdf_for_cdo_no_time_dim_var(air_temperature):
    """Test to_compressed_netcdf if `for_cdo=True` and one var without `time_dim`."""
    ds = air_temperature
    ds["air_mean"] = ds["air"].isel(time=0)
    ds.to_compressed_netcdf("./tmp_testdir/test.nc", for_cdo=True)
    os.remove("./tmp_testdir/test.nc")


def get_zarr_size(fn):
    """Get size of zarr file excluding metadata"""
    # Open file
    grp = zarr.open_group(fn)
    # Collect size
    total = 0
    for var in list(grp.keys()):
        nbytes_stored = grp[var].nbytes_stored
        if isinstance(nbytes_stored, int):
            total += nbytes_stored
        else:
            total += nbytes_stored()
    return total


def test_to_compressed_zarr(rasm):
    """Test to_compressed_zarr reduces size on disk."""
    ds = rasm
    label = "file"
    # save
    encoding = {
        var: {compressor_key: None} for var in ds.data_vars
    }  # deactivate default compression
    ds.to_zarr(f"./tmp_testdir/{label}.zarr", mode="w", encoding=encoding)
    ds.to_compressed_zarr(f"./tmp_testdir/{label}_compressed.zarr", mode="w")
    # check size reduction
    ori_size = get_zarr_size(f"./tmp_testdir/{label}.zarr")
    compressed_size = get_zarr_size(f"./tmp_testdir/{label}_compressed.zarr")
    assert compressed_size < ori_size
    shutil.rmtree(f"./tmp_testdir/{label}.zarr")
    shutil.rmtree(f"./tmp_testdir/{label}_compressed.zarr")


def test_to_compressed_zarr_individual_compressors(eraint_uvz):
    """Test non-default compressors."""
    ds = eraint_uvz
    label = "file"
    # save
    encoding = {
        var: {compressor_key: None} for var in ds.data_vars
    }  # deactivate default compression
    ds.to_zarr(f"./tmp_testdir/{label}.zarr", mode="w", encoding=encoding)
    compressors = {
        "u": BloscCodec(cname="zstd", clevel=7),
        "v": BloscCodec(cname="zlib", clevel=7, shuffle=BloscShuffle.bitshuffle),
    }
    ds.to_compressed_zarr(
        f"./tmp_testdir/{label}_compressed.zarr", compressors, mode="w"
    )
    # check size reduction
    ori_size = get_zarr_size(f"./tmp_testdir/{label}.zarr")
    compressed_size = get_zarr_size(f"./tmp_testdir/{label}_compressed.zarr")
    assert compressed_size < ori_size
    shutil.rmtree(f"./tmp_testdir/{label}.zarr")
    shutil.rmtree(f"./tmp_testdir/{label}_compressed.zarr")



import matplotlib.pyplot as plt
import pytest
import xarray as xr

import xbitinfo as xb
from xbitinfo.graphics import add_bitinfo_labels, plot_bitinformation


def test_add_bitinfo_labels():
    ds = xr.tutorial.load_dataset("air_temperature")
    info_per_bit = xb.get_bitinformation(ds, dim="lon")
    inflevels = [1.0, 0.9999, 0.99, 0.975, 0.95]
    keepbits = [23, 14, 7, 6, 5]
    ds_bitrounded_along_lon = xb.bitround.bitround_along_dim(
        ds, info_per_bit, dim="lon", inflevels=inflevels
    )
    diff = (ds - ds_bitrounded_along_lon)["air"].isel(time=0)
    ax = plt.gca()
    diff.plot()

    with pytest.raises(KeyError):
        add_bitinfo_labels(diff, info_per_bit, inflevels, keepbits)

    add_bitinfo_labels(diff, info_per_bit, inflevels)

    # Check if a Matplotlib figure object is created
    assert plt.gcf() is not None

    # Check if the text labels were added to the plot
    assert len(ax.texts) == 2 * len(inflevels)

    # Check if the plot contains the expected number of lines
    assert len(ax.lines) == len(inflevels)

    # Check if the labels have the correct content
    if inflevels is None:
        expected_inflevels = ["100.0%", "100.0%", "99.88%", "98.89%", "95.28%"]
        for i, keep in enumerate(keepbits):
            inf_text = expected_inflevels[i]
            keepbits_text = f"keepbits = {keep}"
            assert ax.texts[i].get_text() == inf_text
            assert ax.texts[i + 5].get_text() == keepbits_text

    if keepbits is None:
        expected_keepbits = keepbits
        for i, inf in enumerate(inflevels):
            inf_text = str(round(inf * 100, 2)) + "%"
            keepbits_text = expected_keepbits[i]
            assert ax.texts[i].get_text() == inf_text
            assert ax.texts[i + 5].get_text() == keepbits_text
    # Cleanup the plot
    plt.close()


@pytest.mark.parametrize("dtype", ["float64", "float32", "float16"])
def test_plot_bitinformation(dtype):
    rasm = xr.tutorial.load_dataset("air_temperature")
    ds = rasm.astype(dtype)
    info_per_bit = xb.get_bitinformation(ds, dim="lon")
    plot_bitinformation(info_per_bit)



import dask.array as da
import numpy as np
import numpy.ma as nm


def exponent_bias(dtype):
    """
    Returns the exponent bias for a given floating-point dtype.

    Example
    -------
    >>> exponent_bias("f4")
    127
    >>> exponent_bias("f8")
    1023
    """
    info = np.finfo(dtype)
    exponent_bits = info.bits - info.nmant - 1
    return 2 ** (exponent_bits - 1) - 1


def exponent_mask(dtype):
    """
    Returns exponent mask for a given floating-point dtype.

    Example
    -------
    >>> np.binary_repr(exponent_mask(np.float32), width=32)
    '01111111100000000000000000000000'
    >>> np.binary_repr(exponent_mask(np.float16), width=16)
    '0111110000000000'
    """
    if dtype == np.float16:
        mask = 0x7C00
    elif dtype == np.float32:
        mask = 0x7F80_0000
    elif dtype == np.float64:
        mask = 0x7FF0_0000_0000_0000
    return mask


def signed_exponent(A):
    """
    Transform biased exponent notation to signed exponent notation.

    Parameters
    ----------
    A : :py:class:`numpy.array`
        Array to transform

    Returns
    -------
    B : :py:class:`numpy.array`

    Example
    -------
    >>> A = np.array(0.03125, dtype="float32")
    >>> np.binary_repr(A.view("uint32"), width=32)
    '00111101000000000000000000000000'
    >>> np.binary_repr(signed_exponent(A), width=32)
    '01000010100000000000000000000000'
    >>> A = np.array(0.03125, dtype="float64")
    >>> np.binary_repr(A.view("uint64"), width=64)
    '0011111110100000000000000000000000000000000000000000000000000000'
    >>> np.binary_repr(signed_exponent(A), width=64)
    '0100000001010000000000000000000000000000000000000000000000000000'
    """
    itemsize = A.dtype.itemsize
    uinttype = f"u{itemsize}"
    inttype = f"i{itemsize}"

    sign_mask = 1 << np.finfo(A.dtype).bits - 1
    sfmask = sign_mask | (1 << np.finfo(A.dtype).nmant) - 1
    emask = exponent_mask(A.dtype)
    esignmask = sign_mask >> 1

    sbits = np.finfo(A.dtype).nmant
    if itemsize == 8:
        sbits = np.uint64(sbits)
        emask = np.uint64(emask)
    bias = exponent_bias(A.dtype)

    ui = A.view(uinttype)
    sf = ui & sfmask
    e = ((ui & emask) >> sbits).astype(inttype) - bias
    max_eabs = np.iinfo(A.view(uinttype).dtype).max >> sbits
    eabs = abs(e) % (max_eabs + 1)
    esign = np.where(e < 0, esignmask, 0)
    if itemsize == 8:
        eabs = np.uint64(eabs)
        esign = np.uint64(esign)
    esigned = esign | (eabs << sbits)
    B = (sf | esigned).view(np.int64)
    return B


def bitpaircount_u1(a, b):
    assert a.dtype == "u1"
    assert b.dtype == "u1"
    unpack_a = (
        a.flatten()
        .map_blocks(
            np.unpackbits,
            drop_axis=0,
            meta=np.array((), dtype=np.uint8),
            chunks=(a.size * 8,),
        )
        .astype("u1")
    )
    unpack_b = (
        b.flatten()
        .map_blocks(
            np.unpackbits,
            drop_axis=0,
            meta=np.array((), dtype=np.uint8),
            chunks=(b.size * 8,),
        )
        .astype("u1")
    )
    index = ((unpack_a << 1) | unpack_b).reshape(-1, 8)

    selection = np.array([0, 1, 2, 3], dtype="u1")
    sel = np.where((index[..., np.newaxis]) == selection, True, False)
    to_return = sel.sum(axis=0).reshape(8, 2, 2)
    return to_return


def bitpaircount(a, b):
    assert a.dtype.kind == "u"
    assert b.dtype.kind == "u"
    nbytes = max(a.dtype.itemsize, b.dtype.itemsize)

    a, b = np.broadcast_arrays(a, b)

    bytewise_counts = []
    for i in range(nbytes):
        s = (nbytes - 1 - i) * 8
        bitc = bitpaircount_u1((a >> s).astype("u1"), (b >> s).astype("u1"))
        bytewise_counts.append(bitc)
    return np.concatenate(bytewise_counts, axis=0)


def mutual_information(a, b, base=2):
    size = np.prod(np.broadcast_shapes(a.shape, b.shape))
    counts = bitpaircount(a, b)

    p = counts.astype("float") / size
    p = da.ma.masked_equal(p, 0)
    pr = p.sum(axis=-1)[..., np.newaxis]
    ps = p.sum(axis=-2)[..., np.newaxis, :]
    mutual_info = (p * np.ma.log(p / (pr * ps))).sum(axis=(-1, -2)) / np.log(base)
    return mutual_info


def bitinformation(a, axis=0):
    sa = tuple(slice(0, -1) if i == axis else slice(None) for i in range(len(a.shape)))
    sb = tuple(
        slice(1, None) if i == axis else slice(None) for i in range(len(a.shape))
    )
    return mutual_information(a[sa], a[sb])



"""Tests for `xbitinfo.get_keepbits`."""

import pytest
import xarray as xr

import xbitinfo as xb


@pytest.fixture
def rasm_info_per_bit(rasm):
    return xb.get_bitinformation(rasm, axis=0)


def test_get_keepbits(rasm_info_per_bit):
    """Test xb.get_keepbits returns xr.Dataset."""
    assert isinstance(xb.get_keepbits(rasm_info_per_bit), xr.Dataset)


def test_get_keepbits_inflevel_1(rasm_info_per_bit):
    """Test xb.get_keepbits returns all mantissa bits for inflevel 1."""
    keepbits = xb.get_keepbits(rasm_info_per_bit, inflevel=1)
    assert (keepbits == 53).all()


@pytest.mark.parametrize("inflevel", [0.99, [0.99, 1.0]])
def test_get_keepbits_inflevel_dim(rasm_info_per_bit, inflevel):
    """Test xb.get_keepbits returns inflevel as dim."""
    keepbits = xb.get_keepbits(rasm_info_per_bit, inflevel=inflevel)
    assert "inflevel" in keepbits.dims
    if isinstance(inflevel, (int, float)):
        inflevel = [inflevel]
    assert (keepbits.inflevel == inflevel).all()


def test_get_keepbits_informationFilter():
    """
    Test the `get_keepbits` function with different information filters.

    This test function checks the behavior of the `get_keepbits` function when applying gradient information filter.
    The dataset contains artificial information and thus applying the filter should result in lesser number of bits
    than what should be when filter is None.


    Raises:
        AssertionError: If the test conditions are not met.

    """

    bit32_values = [
        "±",
        "e1",
        "e2",
        "e3",
        "e4",
        "e5",
        "e6",
        "e7",
        "e8",
        "m1",
        "m2",
        "m3",
        "m4",
        "m5",
        "m6",
        "m7",
        "m8",
        "m9",
        "m10",
        "m11",
        "m12",
        "m13",
        "m14",
        "m15",
        "m16",
        "m17",
        "m18",
        "m19",
        "m20",
        "m21",
        "m22",
        "m23",
    ]
    data_variable = xr.DataArray(
        data=[
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            1.11799129e-01,
            8.19114977e-01,
            4.41578500e-01,
            3.25470303e-01,
            4.35195738e-01,
            2.81462993e-01,
            2.10719742e-01,
            1.46638224e-01,
            9.24534031e-02,
            4.41090879e-02,
            1.13842504e-02,
            8.20088050e-04,
            2.62239097e-06,
            7.11284508e-07,
            1.18183485e-06,
            9.49338973e-09,
            1.80859255e-07,
            7.72662891e-07,
            1.37865391e-05,
            2.11117224e-06,
            2.01353088e-07,
            3.20755770e-02,
            6.06721012e-03,
            2.25987148e-04,
            1.71530452e-06,
            5.13067595e-03,
        ],
        coords={"bitfloat32": bit32_values, "dim": "x"},
        dims=["bitfloat32"],
    )
    info_ds = xr.Dataset({"RH2": data_variable})
    Keepbits_FilterNone = xb.get_keepbits(
        info_ds,
        inflevel=[0.99],
        information_filter=None,
        **{"threshold": 0.7, "tolerance": 0.001},
    )
    Keepbits_FilterNone_Value = Keepbits_FilterNone["RH2"].values
    assert Keepbits_FilterNone_Value == 19

    Keepbits_FilterGradient = xb.get_keepbits(
        info_ds,
        inflevel=[0.99],
        information_filter="Gradient",
        **{"threshold": 0.7, "tolerance": 0.001},
    )
    Keepbits_FilterGradient_Value = Keepbits_FilterGradient["RH2"].values
    assert Keepbits_FilterGradient_Value == 7


def test_get_keepbits_informationFilter_1():
    """
    Test the `get_keepbits` function with different information filters.

    This test function checks the behavior of the `get_keepbits` function when applying gradient information filter.
    The dataset does not contain artificial information and thus the number of keepbits when gradient filter is applied
    should be equal to when filter is None.

    Raises:
        AssertionError: If the test conditions are not met.

    """

    ds = xr.tutorial.load_dataset("air_temperature")
    info = xb.get_bitinformation(ds, dim="lat")
    Keepbits_FilterNone = xb.get_keepbits(
        info,
        inflevel=[0.99],
        information_filter=None,
        **{"threshold": 0.7, "tolerance": 0.001},
    )
    Keepbits_FilterNone_Value = Keepbits_FilterNone["air"].values

    Keepbits_FilterGradient = xb.get_keepbits(
        info,
        inflevel=[0.99],
        information_filter="Gradient",
        **{"threshold": 0.7, "tolerance": 0.001},
    )

    Keepbits_FilterGradient_Value = Keepbits_FilterGradient["air"].values
    assert Keepbits_FilterNone_Value == Keepbits_FilterGradient_Value



import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .xbitinfo import _cdf_from_info_per_bit, bit_partitioning, get_keepbits


def add_bitinfo_labels(
    da,
    info_per_bit,
    inflevels=None,
    keepbits=None,
    ax=None,
    x_dim_name="lon",
    y_dim_name="lat",
    lon_coord_name="guess",
    lat_coord_name="guess",
    label_latitude="center",
    label_latitude_offset=8,
    **kwargs,
):
    """
    Helper function for visualization of Figure 3 in Klöwer et al. 2021.
    Adds latitudinal lines and labels with keepbits and information content for each slice.

    Klöwer, M., Razinger, M., Dominguez, J. J., Düben, P. D., & Palmer, T. N. (2021).
    Compressing atmospheric data into its real information content. Nature Computational Science, 1(11), 713–724. doi: 10/gnm4jj

    Parameters
    ----------
    da : :py:func:`xarray.DataArray`
      Plotted data
    info_per_bit : dict
      Information content of each bit for each variable in ``da``. This is the output from :py:func:`xbitinfo.xbitinfo.get_bitinformation`.
    inflevels : list of floats
      Level of information that shall be preserved.
    ax : plt.Axes or None
      Axes. If ``None``, get current axis.
    x_dim_name : str
      Name of the x dimension. Defaults to ``"lon"``.
    y_dim_name : str
      Name of the y dimension. Defaults to ``"lat"``.
    lon_coord_name : str
      Name of the longitude coordinate. Only matters when plotting with multi-dimensional coordinates (i.e. curvilinear grids) with ``cartopy`` (when ``transform=ccrs.Geodetic()`` must be also set via ``kwargs``). Defaults to ``x_dim_name``.
    lat_coord_name : str
      Name of the latitude coordinate. Only matters when plotting with multi-dimensional coordinates (i.e. curvilinear grids) with ``cartopy`` (when ``transform=ccrs.Geodetic()`` must be also set via ``kwargs``). Defaults to ``y_dim_name``.
    label_latitude :  float or str
      Latitude for the label. Defaults to ``"center"``, which uses the mean ``lat_coord_name``.
    label_latitude_offset : float
      Distance between ``keepbits = int`` and ``x%`` label. Defaults to ``8``.
    kwargs : dict
      Kwargs to be passed to ``ax.text`` and ``ax.plot``. Use ``transform=ccrs.Geodetic()`` when using ``cartopy``

    Returns
    -------

    Example
    -------
    Plotting a single-dimension coordinate dataset:

    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info_per_bit = xb.get_bitinformation(ds, dim="lon")
    >>> inflevels = [1.0, 0.9999, 0.99, 0.975, 0.95]
    >>> keepbits = None
    >>> ds_bitrounded_along_lon = xb.bitround.bitround_along_dim(
    ...     ds, info_per_bit, dim="lon", inflevels=inflevels
    ... )
    >>> diff = (ds - ds_bitrounded_along_lon)["air"].isel(time=0)
    >>> diff.plot()  # doctest: +ELLIPSIS
    <matplotlib.collections.QuadMesh object at ...>
    >>> add_bitinfo_labels(
    ...     diff, info_per_bit, inflevels, keepbits
    ... )  # doctest: +ELLIPSIS

    Plotting a multi-dimensional coordinate dataset

    >>> v = "Tair"
    >>> ds = xr.tutorial.load_dataset("rasm")
    >>> dim = "y"
    >>> info_per_bit = xb.get_bitinformation(ds, dim=dim)
    >>> ds_bitrounded_along_lon = xb.bitround.bitround_along_dim(
    ...     ds, info_per_bit, dim=dim, inflevels=inflevels
    ... )
    >>> import cartopy.crs as ccrs  # doctest: +SKIP
    >>> fig, axis = plt.subplots(  # doctest: +SKIP
    ...     1, 1, subplot_kw=dict(projection=ccrs.PlateCarree())
    ... )
    >>> (ds - ds_bitrounded_along_lon)[v].isel(time=-10).plot(
    ...     ax=axis, transform=ccrs.PlateCarree()
    ... )  # doctest: +SKIP
    >>> add_bitinfo_labels(
    ...     (ds - ds_bitrounded_along_lon)[v].isel(time=0),
    ...     lon_coord_name="xc",
    ...     lat_coord_name="yc",
    ...     x_dim_name="x",
    ...     y_dim_name="y",
    ...     transform=ccrs.Geodetic(),
    ... )  # doctest: +SKIP

    """

    if inflevels is None and keepbits is None:
        raise KeyError("Either inflevels or keepbits need to be provided")
    elif inflevels is not None and keepbits is not None:
        raise KeyError("Only inflevels or keepbits can be provided")
    if lon_coord_name == "guess":
        lon_coord_name = x_dim_name
    if lat_coord_name == "guess":
        lat_coord_name = y_dim_name
    if label_latitude == "center":
        label_latitude = da[lat_coord_name].mean()
    if ax is None:
        ax = plt.gca()

    dimension_dict = info_per_bit.dims
    dimension_list = list(dimension_dict.keys())
    dimension = dimension_list[0]
    CDF = _cdf_from_info_per_bit(info_per_bit, dimension)
    CDF_DataArray = CDF[da.name]

    data_type = np.dtype(dimension.replace("bit", ""))
    _, _, n_exp, _ = bit_partitioning(data_type)
    if inflevels is None:
        inflevels = []
        for i, keep in enumerate(keepbits):
            mantissa_index = keep + n_exp
            inflevels.append(CDF_DataArray[mantissa_index].values)

    if keepbits is None:
        keepbits = [get_keepbits(info_per_bit, ilev) for ilev in inflevels]

    if isinstance(keepbits, list) and all(
        isinstance(ds, xr.Dataset) for ds in keepbits
    ):
        keepbits_data = []
        for ds in keepbits:
            data_var = ds[da.name].values
            for value in data_var:
                keepbits_data.append(value)
        keepbits = keepbits_data

    stride = da[x_dim_name].size // len(inflevels)

    for i, inf in enumerate(inflevels):
        # draw latitude line
        lons = da.isel({x_dim_name: stride * i})[lon_coord_name]
        lats = da.isel({x_dim_name: stride * i})[lat_coord_name]
        lons, lats = xr.broadcast(lons, lats)
        ax.plot(lons, lats, color="k", linewidth=1, **kwargs)

        # write inflevel
        t = ax.text(
            da.isel(
                {
                    x_dim_name: int(stride * (i + 0.5)),
                    y_dim_name: da[y_dim_name].size // 2,
                }
            )[lon_coord_name].values,
            label_latitude - label_latitude_offset,
            str(round(inf * 100, 2)) + "%",
            horizontalalignment="center",
            color="k",
            **kwargs,
        )
        t.set_bbox(dict(facecolor="white", alpha=0.9, edgecolor="white"))

    for i, keep in enumerate(keepbits):
        # write keepbits
        t_keepbits = ax.text(
            da.isel(
                {
                    x_dim_name: int(stride * (i + 0.5)),
                    y_dim_name: da[y_dim_name].size // 2,
                }
            )[lon_coord_name].values,
            label_latitude + label_latitude_offset,
            f"keepbits = {keep}",
            horizontalalignment="center",
            color="k",
            **kwargs,
        )
        t_keepbits.set_bbox(dict(facecolor="white", alpha=0.9, edgecolor="white"))


def split_dataset_by_dims(info_per_bit):
    """Split dataset by its dimensions.

    Parameters
    ----------
    info_per_bit : dict
      Information content of each bit for each variable in ``da``. This is the output from :py:func:`xbitinfo.xbitinfo.get_bitinformation`.

    Returns
    -------
    var_by_dim : dict
      Dictionary containing the dimensions of the datasets as keys and the dataset using the dimension as value.
    """
    var_by_dim = {d: [] for d in info_per_bit.dims}
    for var in info_per_bit.data_vars:
        assert (
            len(info_per_bit[var].dims) == 1
        ), f"Variable {var} has more than one dimension."
        var_by_dim[info_per_bit[var].dims[0]].append(var)
    return var_by_dim


def plot_bitinformation(bitinfo, information_filter=None, cmap="turku", crop=None):
    """Plot bitwise information content as in Klöwer et al. 2021 Figure 2.

    Klöwer, M., Razinger, M., Dominguez, J. J., Düben, P. D., & Palmer, T. N. (2021).
    Compressing atmospheric data into its real information content.
    Nature Computational Science, 1(11), 713–724. doi: 10/gnm4jj

    Parameters
    ----------
    bitinfo : :py:func:`xarray.Dataset`
      Containing the bitwise information content for each variable
    information_filter : str
      Filter algorithm to filter artificial information content. Defaults to ``None``.
      Available filters are: ``"Gradient"``.
    cmap : str or plt.cm
      Colormap. Defaults to ``"turku"``.
    crop : int
      Maximum bits to show in figure.

    Returns
    -------
    fig : matplotlib figure

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("air_temperature")
    >>> info_per_bit = xb.get_bitinformation(ds, dim="lon")
    >>> xb.plot_bitinformation(info_per_bit)
    <Figure size 1200x800 with 4 Axes>

    """

    bitinfo = bitinfo.squeeze()
    assert (
        "dim" not in bitinfo.dims
    ), "Found dependence of bitinformation on dimension. Please reduce data first by e.g. `bitinfo.max(dim='dim')`"
    vars_by_dim = split_dataset_by_dims(bitinfo)
    bitinfo_all = bitinfo
    subfigure_data = [None] * len(vars_by_dim)
    for d, (dim, vars) in enumerate(vars_by_dim.items()):
        bitinfo = bitinfo_all[vars]
        data_type = np.dtype(dim.replace("bit", ""))
        n_bits, n_sign, n_exp, n_mant = bit_partitioning(data_type)
        nonmantissa_bits = n_bits - n_mant
        if crop is None:
            bits_to_show = n_bits
        else:
            bits_to_show = int(np.min([crop, n_bits]))
        nvars = len(bitinfo)
        varnames = list(bitinfo.keys())

        if information_filter == "Gradient":
            infbits_dict = get_keepbits(
                bitinfo,
                0.99,
                information_filter,
                **{"threshold": 0.7, "tolerance": 0.001},
            )
            infbits100_dict = get_keepbits(
                bitinfo,
                0.999999999,
                information_filter,
                **{"threshold": 0.7, "tolerance": 0.001},
            )
        else:
            infbits_dict = get_keepbits(bitinfo, 0.99)
            infbits100_dict = get_keepbits(bitinfo, 0.999999999)

        ICnan = np.zeros((nvars, 64))
        infbits = np.zeros(nvars)
        infbits100 = np.zeros(nvars)
        ICnan[:, :] = np.nan
        for v, var in enumerate(varnames):
            ic = bitinfo[var].squeeze(drop=True)
            ICnan[v, : len(ic)] = ic
            # infbits are all bits, infbits_dict were mantissa bits
            infbits[v] = infbits_dict[var] + nonmantissa_bits
            infbits100[v] = infbits100_dict[var] + nonmantissa_bits
        ICnan = np.where(ICnan == 0, np.nan, ICnan)
        ICcsum = np.nancumsum(ICnan, axis=1)

        infbitsy = np.hstack([0, np.repeat(np.arange(1, nvars), 2), nvars])
        infbitsx = np.repeat(infbits, 2)
        infbitsx100 = np.repeat(infbits100, 2)

        fig_height = np.max([4, 4 + (nvars - 10) * 0.2])  # auto adjust to nvars

        subfigure_data[d] = {}
        subfigure_data[d]["fig_height"] = fig_height
        subfigure_data[d]["nvars"] = nvars
        subfigure_data[d]["varnames"] = varnames
        subfigure_data[d]["ICnan"] = ICnan
        subfigure_data[d]["ICcsum"] = ICcsum
        subfigure_data[d]["infbits"] = infbits
        subfigure_data[d]["infbitsx"] = infbitsx
        subfigure_data[d]["infbitsy"] = infbitsy
        subfigure_data[d]["infbitsx100"] = infbitsx100
        subfigure_data[d]["nbits"] = (n_sign, n_exp, n_bits, n_mant, nonmantissa_bits)
        subfigure_data[d]["bits_to_show"] = bits_to_show

    fig_heights = [subfig["fig_height"] for subfig in subfigure_data]
    fig = plt.figure(figsize=(12, sum(fig_heights) + 2 * 2))
    fig_heights_incl_cax = fig_heights + [2 / (sum(fig_heights) + 2)] * 2
    grid = fig.add_gridspec(
        len(subfigure_data) + 2, 1, height_ratios=fig_heights_incl_cax
    )

    axs = []
    for i in range(len(subfigure_data) + 2):
        ax = fig.add_subplot(grid[i, 0])
        axs.append(ax)

    if isinstance(axs, plt.Axes):
        axs = [axs]

    fig.suptitle(
        "Real bitwise information content",
        x=0.05,
        y=0.98,
        fontweight="bold",
        horizontalalignment="left",
    )

    if cmap == "turku":
        import cmcrameri.cm as cmc

        cmap = cmc.turku_r

    max_bits_to_show = np.max([d["bits_to_show"] for d in subfigure_data])

    for d, subfig in enumerate(subfigure_data):
        infbits = subfig["infbits"]
        nvars = subfig["nvars"]
        n_sign, n_exp, n_bits, n_mant, nonmantissa_bits = subfig["nbits"]
        ICcsum = subfig["ICcsum"]
        ICnan = subfig["ICnan"]
        infbitsy = subfig["infbitsy"]
        infbitsx = subfig["infbitsx"]
        infbitsx100 = subfig["infbitsx100"]
        varnames = subfig["varnames"]
        bits_to_show = subfig["bits_to_show"]

        mbits_to_show = bits_to_show - nonmantissa_bits

        axs[d].invert_yaxis()
        axs[d].set_box_aspect(1 / max_bits_to_show * nvars)

        ax1right = axs[d].twinx()
        ax1right.invert_yaxis()
        ax1right.set_box_aspect(1 / max_bits_to_show * nvars)

        pcm = axs[d].pcolormesh(ICnan, vmin=0, vmax=1, cmap=cmap)

        if d == len(subfigure_data) - 1:
            cax = axs[len(subfigure_data)]
            lax = axs[len(subfigure_data) + 1]
            lax.axis("off")
            cbar = plt.colorbar(pcm, cax=cax, orientation="horizontal")
            cbar.set_label("information content [bit]")

        # 99% of real information enclosed
        l0 = axs[d].plot(
            np.hstack([infbits, infbits[-1]]),
            np.arange(nvars + 1),
            "C1",
            ds="steps-pre",
            zorder=10,
            label="99% of\ninformation",
        )

        # grey shading
        axs[d].fill_betweenx(
            infbitsy,
            infbitsx,
            np.ones(len(infbitsx)) * bits_to_show,
            alpha=0.4,
            color="grey",
        )
        axs[d].fill_betweenx(
            infbitsy,
            infbitsx100,
            np.ones(len(infbitsx)) * bits_to_show,
            alpha=0.1,
            color="c",
        )
        axs[d].fill_betweenx(
            infbitsy,
            infbitsx100,
            np.ones(len(infbitsx)) * bits_to_show,
            alpha=0.3,
            facecolor="none",
            edgecolor="c",
        )

        # for legend only
        l1 = axs[d].fill_betweenx(
            [-1, -1],
            [-1, -1],
            [-1, -1],
            color="burlywood",
            label="last 1% of\ninformation",
            alpha=0.5,
        )
        l2 = axs[d].fill_betweenx(
            [-1, -1],
            [-1, -1],
            [-1, -1],
            facecolor="teal",
            edgecolor="c",
            label="false information",
            alpha=0.3,
        )
        l3 = axs[d].fill_betweenx(
            [-1, -1], [-1, -1], [-1, -1], color="w", label="unused bits", edgecolor="k"
        )

        if n_sign > 0:
            axs[d].axvline(n_sign, color="k", lw=1, zorder=3)
        axs[d].axvline(nonmantissa_bits, color="k", lw=1, zorder=3)

        axs[d].set_ylim(nvars, 0)
        ax1right.set_ylim(nvars, 0)

        axs[d].set_yticks(np.arange(nvars) + 0.5)
        ax1right.set_yticks(np.arange(nvars) + 0.5)
        axs[d].set_yticklabels(varnames)
        ax1right.set_yticklabels([f"{i:4.1f}" for i in ICcsum[:, -1]])
        if d == len(subfigure_data) // 2:
            ax1right.set_ylabel("total information\nper value [bit]")

        axs[d].text(
            infbits[0] + 0.1,
            0.8,
            f"{int(infbits[0]-nonmantissa_bits)} mantissa bits",
            fontsize=8,
            color="saddlebrown",
        )
        for i in range(1, nvars):
            axs[d].text(
                infbits[i] + 0.1,
                (i) + 0.8,
                f"{int(infbits[i]-9)}",
                fontsize=8,
                color="saddlebrown",
            )

        major_xticks = np.array([n_sign, n_sign + n_exp, n_bits], dtype="int")
        axs[d].set_xticks(major_xticks[major_xticks <= bits_to_show])
        minor_xticks = np.hstack(
            [
                np.arange(n_sign, nonmantissa_bits - 1),
                np.arange(nonmantissa_bits, n_bits - 1),
            ]
        )
        axs[d].set_xticks(
            minor_xticks[minor_xticks <= bits_to_show],
            minor=True,
        )
        axs[d].set_xticklabels([])
        if n_sign > 0:
            axs[d].text(0, nvars + 1.2, "sign", rotation=90)
        if n_exp > 0:
            axs[d].text(
                n_sign + n_exp / 2,
                nvars + 1.2,
                "exponent bits",
                color="darkslategrey",
                horizontalalignment="center",
                verticalalignment="center",
            )
        axs[d].text(
            n_sign + n_exp + mbits_to_show / 2,
            nvars + 1.2,
            "mantissa bits",
            horizontalalignment="center",
            verticalalignment="center",
        )

        # Set xticklabels
        ## Set exponent labels
        for e, i in enumerate(range(n_sign, np.min([n_sign + n_exp, bits_to_show]))):
            axs[d].text(
                i + 0.5,
                nvars + 0.5,
                e + 1,
                ha="center",
                fontsize=7,
                color="darkslategrey",
            )
        ## Set mantissa labels
        for m, i in enumerate(
            range(n_sign + n_exp, np.min([n_sign + n_exp + n_mant, bits_to_show]))
        ):
            axs[d].text(i + 0.5, nvars + 0.5, m + 1, ha="center", fontsize=7)

        if d == len(subfigure_data) - 1:
            lax.legend(
                bbox_to_anchor=(0.5, 0),
                loc="center",
                framealpha=0.6,
                ncol=4,
                handles=[l0[0], l1, l2, l3],
            )
        axs[d].set_xlim(0, bits_to_show)

    fig.show()

    return fig


def plot_distribution(ds, nbins=1000, cmap="viridis", offset=0.01, close_zero=1e-2):
    """Plot statistical distributions of all variables as in Klöwer et al. 2021 Figure SI 1.
    For large data subsetting, i.e. ds = ds.isel(x=slice(None, None, 100)) is recommended.

    Klöwer, M., Razinger, M., Dominguez, J. J., Düben, P. D., & Palmer, T. N. (2021).
    Compressing atmospheric data into its real information content.
    Nature Computational Science, 1(11), 713–724. doi: 10/gnm4jj

    Parameters
    ----------
    bitinfo : :py:class:`xarray.Dataset`
      Raw input values for distributions
    nbints : int
      Number of bins for histograms across all variable range. Defaults to ``1000``.
    cmap : str
      Which matplotlib colormap to use. Defaults to ``"viridis"``.
    offset : float
      Offset on the yaxis between variables 0 lines. Defaults to ``0.01``.
    close_zero : float
      Threshold where to stop close to 0, when distributions ranges from negative to positive.
      Increase this value when seeing an unexpected dip around 0 in the distribution. Defaults to ``0.01``.

    Returns
    -------
    fig : matplotlib figure

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("eraint_uvz")
    >>> xb.plot_distribution(ds)
    <Axes: title={'center': 'Statistical distributions'}, xlabel='value', ylabel='Probability density'>

    """

    if not isinstance(ds, xr.Dataset):
        raise ValueError(
            f"plot_distribution(ds), requires xr.Dataset, found {type(ds)}"
        )

    varnames = list(ds.data_vars)
    nvars = len(varnames)
    ds = ds[varnames].squeeze()
    gmin, gmax = ds.to_array().min(), ds.to_array().max()
    f = 2  # factor for bounds
    if gmin < 0 and gmax > 0:
        bins_neg = np.geomspace(gmin * f, -close_zero, nbins // 2 + 1, dtype=float)
        bins_pos = np.geomspace(close_zero, gmax * f, nbins // 2, dtype=float)
        bins = np.concatenate([bins_neg, bins_pos], axis=-1)
    else:
        bins = np.geomspace(gmin / f, gmax * f, nbins + 1, dtype=float)

    H = np.zeros((nvars, nbins))
    for i, v in enumerate(varnames):
        d = ds[v].data.flatten()
        d = d[~np.isnan(d)]  # drop NaN
        H[i, :], _ = np.histogram(d, bins=bins, density=True)
        H[i, :] = H[i, :] / np.sum(H[i, :])  # normalize

    fig, ax = plt.subplots(1, 1, figsize=(5, 2 + nvars / 10))
    cmap_instance = plt.get_cmap(cmap, nvars)

    for i in range(nvars):
        c = cmap_instance.colors[i]
        plt.plot(bins[:-1], H[i, :] + offset * i, color=c)
        plt.fill_between(
            bins[:-1], H[i, :] + offset * i, offset * i, alpha=0.5, color=c
        )
    ax.set_xscale(
        "symlog"
    )  # https://stackoverflow.com/questions/43372499/plot-negative-values-on-a-log-scale
    ymax = max(0.05, nvars / 100 + 0.02)  # at least 10% y
    ax.set_ylim([-offset / 2, ymax])
    ax.set_xlim([bins[0], bins[-1]])
    minyticks = np.arange(0, ymax + 0.01, offset)
    majyticks = np.arange(0, ymax + 0.01, offset * 5)
    ax.set_yticks(minyticks, minor=True)
    ax.set_yticks(majyticks, minor=False)
    ax.set_yticklabels([str(int(i * 100)) + "%" for i in majyticks])

    axright = ax.twinx()
    axright.set_ylim([-offset / 2, ymax])
    axright.set_yticks(minyticks, minor=False)
    axright.set_yticklabels(
        varnames + [""] * (len(minyticks) - len(varnames)), minor=False
    )
    ax.set_xlabel("value")
    ax.set_ylabel("Probability density")
    ax.set_title("Statistical distributions")
    return ax



"""Tests for `xbitinfo` package."""

import os
import warnings

import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_allclose, assert_equal
from xarray.core import formatting
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset
from xarray.core.variable import Variable
from xarray.testing import assert_identical

import xbitinfo as xb


def assert_different(a, b):
    """Raises an AssertionError if two objects are equal. This will match
    data values, dimensions and coordinates, but not names or attributes
    (except for Dataset objects for which the variable names must match).
    Arrays with NaN in the same location are considered equal.
    Parameters
    ----------
    a : xarray.Dataset, xarray.DataArray or xarray.Variable
        The first object to compare.
    b : xarray.Dataset, xarray.DataArray or xarray.Variable
        The second object to compare.
    See Also
    --------
    assert_identical, assert_allclose, Dataset.equals, DataArray.equals
    numpy.testing.assert_array_equal
    """
    __tracebackhide__ = True
    assert isinstance(a, type(b))
    if isinstance(a, (Variable, DataArray)):
        assert not a.equals(b), formatting.diff_array_repr(a, b, "equals")
    elif isinstance(a, Dataset):
        assert not a.equals(b), formatting.diff_dataset_repr(a, b, "equals")
    else:
        raise TypeError(f"{type(a)} not supported by assertion comparison")


def bitinfo_assert_equal(bitinfo1, bitinfo2):
    assert list(bitinfo1.keys()) == list(bitinfo2.keys()), print(
        f"lhs = {bitinfo1.keys()} vs rhs = {bitinfo2.keys()}"
    )
    for v in bitinfo1.keys():
        assert_equal(bitinfo1[v], bitinfo2[v])


def bitinfo_assert_allclose(bitinfo1, bitinfo2, **kwargs):
    assert list(bitinfo1.keys()) == list(bitinfo2.keys()), print(
        f"lhs = {bitinfo1.keys()} vs rhs = {bitinfo2.keys()}"
    )
    for v in bitinfo1.keys():
        assert_allclose(bitinfo1[v], bitinfo2[v], **kwargs)


def bitinfo_assert_different(bitinfo1, bitinfo2):
    """Fail bitinfo different values."""
    assert (bitinfo1 != bitinfo2).any()


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_returns_dataset(implementation):
    """Test xb.get_bitinformation returns xr.Dataset."""
    ds = xr.tutorial.load_dataset("rasm")
    assert isinstance(
        xb.get_bitinformation(ds, implementation=implementation, axis=0), xr.Dataset
    )


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_dim(implementation):
    """Test xb.get_bitinformation is sensitive to dim."""
    ds = xr.tutorial.load_dataset("rasm")
    bitinfo0 = xb.get_bitinformation(ds, axis=0, implementation=implementation)
    bitinfo2 = xb.get_bitinformation(ds, axis=2, implementation=implementation)
    assert_different(bitinfo0, bitinfo2)


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_dim_string_equals_axis_int(implementation):
    """Test xb.get_bitinformation undestands xarray dimension names the same way as axis as integers."""
    ds = xr.tutorial.load_dataset("rasm")
    bitinfox = xb.get_bitinformation(ds, dim="x", implementation=implementation)
    bitinfo2 = xb.get_bitinformation(ds, axis=2, implementation=implementation)
    assert_identical(bitinfox, bitinfo2)


def test_get_bitinformation_masked_value(implementation="julia"):
    """Test xb.get_bitinformation is sensitive to masked_value."""
    ds = xr.tutorial.load_dataset("rasm")
    bitinfo = xb.get_bitinformation(ds, dim="x", implementation=implementation)
    bitinfo_no_mask = xb.get_bitinformation(
        ds, dim="x", masked_value="nothing", implementation=implementation
    )
    bitinfo_no_mask_None = xb.get_bitinformation(
        ds, dim="x", masked_value=None, implementation=implementation
    )
    bitinfo_fillna = xb.get_bitinformation(
        ds.fillna(-999), dim="x", masked_value=-999.0, implementation=implementation
    )
    assert_identical(bitinfo_no_mask, bitinfo_no_mask_None)
    assert_identical(bitinfo, bitinfo_no_mask)
    assert_different(bitinfo, bitinfo_fillna)


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_set_zero_insignificant(implementation):
    """Test xb.get_bitinformation is sensitive to set_zero_insignificant."""
    ds = xr.tutorial.load_dataset("air_temperature")
    dim = "lon"
    bitinfo = xb.get_bitinformation(ds, dim=dim, implementation=implementation)
    bitinfo_szi_False = xb.get_bitinformation(
        ds, dim=dim, set_zero_insignificant=False, implementation=implementation
    )
    try:
        bitinfo_szi_True = xb.get_bitinformation(
            ds, dim=dim, set_zero_insignificant=True, implementation=implementation
        )
        assert_identical(bitinfo, bitinfo_szi_True)
    except NotImplementedError:
        assert implementation == "python"
    if implementation == "python":
        assert_identical(bitinfo, bitinfo_szi_False)
    elif implementation == "julia":
        assert_different(bitinfo, bitinfo_szi_False)


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_confidence(implementation):
    """Test xb.get_bitinformation is sensitive to confidence."""
    ds = xr.tutorial.load_dataset("air_temperature")
    dim = "lon"
    bitinfo = xb.get_bitinformation(ds, dim=dim, implementation=implementation)
    try:
        bitinfo_conf99 = xb.get_bitinformation(
            ds, dim=dim, confidence=0.99, implementation=implementation
        )
        bitinfo_conf50 = xb.get_bitinformation(
            ds, dim=dim, confidence=0.5, implementation=implementation
        )
        assert_different(bitinfo_conf99, bitinfo_conf50)
        assert_identical(bitinfo, bitinfo_conf99)
    except AssertionError:
        assert implementation == "python"


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_label(rasm, implementation):
    """Test xb.get_bitinformation serializes when label given."""
    ds = rasm
    xb.get_bitinformation(
        ds, dim="x", label="./tmp_testdir/rasm", implementation=implementation
    )
    assert os.path.exists("./tmp_testdir/rasm.json")
    # second call should be faster
    xb.get_bitinformation(
        ds, dim="x", label="./tmp_testdir/rasm", implementation=implementation
    )
    os.remove("./tmp_testdir/rasm.json")


@pytest.mark.parametrize("implementation", ["julia", "python"])
@pytest.mark.parametrize("dtype", ["float64", "float32", "float16", "int16"])
def test_get_bitinformation_dtype(rasm, dtype, implementation):
    """Test xb.get_bitinformation returns correct number of bits depending on dtype."""
    dtype = np.dtype(dtype)
    ds = rasm.astype(dtype)
    v = list(ds.data_vars)[0]
    if dtype.kind == "f":
        dtype_bits = np.finfo(dtype).bits
    elif dtype.kind == "i" or dtype.kind == "u":
        dtype_bits = np.iinfo(dtype).bits
    assert (
        len(xb.get_bitinformation(ds, dim="x")[v].coords["bit" + str(dtype)])
        == dtype_bits
    )


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_multidim(rasm, implementation):
    """Test xb.get_bitinformation runs on all dimensions by default"""
    ds = rasm
    bi = xb.get_bitinformation(ds, implementation=implementation)
    # check length of dimension
    assert bi.dims["dim"] == len(ds.dims)
    bi_time = bi.sel(dim="time").Tair.values
    bi_x = bi.sel(dim="x").Tair.values
    bi_y = bi.sel(dim="y").Tair.values
    assert any(bi_time != bi_x)
    assert any(bi_time != bi_y)
    assert any(bi_y != bi_x)


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_different_variables_dims(rasm, implementation):
    """Test xb.get_bitinformation runs with variables of different dimensionality"""
    ds = rasm
    # add variable with different dimensionality
    ds["Tair_mean"] = ds.Tair.mean(dim="time")
    bi = xb.get_bitinformation(ds, implementation=implementation)
    assert all(np.isnan(bi.Tair_mean.sel(dim="time")))
    bi_Tair_mean_x = bi.Tair_mean.sel(dim="x")
    bi_Tair_x = bi.Tair.sel(dim="x")
    assert_different(bi_Tair_mean_x, bi_Tair_x)


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_different_dtypes(rasm, implementation):
    ds = rasm
    ds["Tair32"] = ds.Tair.astype("float32")
    ds["Tair16"] = ds.Tair.astype("float16")
    bi = xb.get_bitinformation(ds, implementation=implementation)
    for bitdim in ["bitfloat16", "bitfloat32", "bitfloat64"]:
        assert bitdim in bi.dims
        assert bitdim in bi.coords


@pytest.mark.parametrize("implementation", ["julia", "python"])
def test_get_bitinformation_dim_list(rasm, implementation):
    bi = xb.get_bitinformation(rasm, dim=["x", "y"], implementation=implementation)
    assert (bi.dim == ["x", "y"]).all()


def test_get_bitinformation_keep_attrs(rasm):
    bi = xb.get_bitinformation(rasm, dim=["x", "y"]).Tair
    assert "units" in bi.attrs
    assert bi.attrs["units"] == 1
    for a in rasm.Tair.attrs.keys():
        assert bi.attrs["source_" + a] == rasm.Tair.attrs[a], print(bi.attrs)


@pytest.mark.parametrize(
    "ds,dim,axis",
    [
        ("ugrid_demo", None, -1),
        ("icon_grid_demo", "ncells", None),
        ("air_temperature", "lon", None),
        ("rasm", "x", None),
        ("ROMS_example", "eta_rho", None),
        ("era52mt", "time", None),
        ("eraint_uvz", "longitude", None),
    ],
)
def test_implementations_agree(ds, dim, axis, request):
    """Test whether the python and julia implementation retrieve the same results"""
    ds = request.getfixturevalue(ds)
    bi_python = xb.get_bitinformation(
        ds,
        dim=dim,
        axis=axis,
        implementation="python",
        set_zero_insignificant=False,
        overwrite=True,
        masked_value=None,
    )
    bi_julia = xb.get_bitinformation(
        ds,
        dim=dim,
        axis=axis,
        implementation="julia",
        set_zero_insignificant=False,
        overwrite=True,
        masked_value=None,
    )
    bitinfo_assert_allclose(bi_python, bi_julia, rtol=1e-4)


@pytest.mark.parametrize("implementation", ["python", "julia"])
@pytest.mark.parametrize("dataset_name", ["air_temperature", "eraint_uvz"])
def test_warn_on_quantized_variables(dataset_name, implementation):
    ds_quantized = xr.tutorial.load_dataset(dataset_name)
    ds_raw = xr.tutorial.load_dataset(dataset_name, mask_and_scale=False)

    with pytest.warns(UserWarning):
        _ = xb.get_bitinformation(ds_quantized, implementation=implementation)

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        _ = xb.get_bitinformation(ds_raw, implementation=implementation)



import numcodecs
import xarray as xr
import zarr
from packaging.version import Version

zarr_version = Version(zarr.__version__)

if zarr_version >= Version("3.0.0"):
    from zarr.codecs import BloscCodec, BloscShuffle

    compressor_key = "compressors"
else:
    from zarr.codecs import Blosc as BloscCodec

    compressor_key = "compressor"

    class BloscShuffle:
        bitshuffle = zarr.codecs.blosc.BITSHUFFLE
        shuffle = zarr.codecs.blosc.SHUFFLE
        noshuffle = zarr.codecs.blosc.NOSHUFFLE


def get_chunksizes(da, for_cdo=False, time_dim="time", chunks=None):
    """Get chunksizes for :py:class:`xarray.DataArray` for ``to_netcdf(encoding)`` from original file.
    If ``for_cdo=True``, ensure ``time_dim`` ``chunksize`` of 1 when compressed."""
    assert isinstance(da, xr.DataArray)
    if chunks:  # use new chunksizes
        return da.chunk(chunks).data.chunksize
    if for_cdo:  # take shape as chunksize and ensure time chunksize 1
        if time_dim in da.dims:
            time_axis_num = da.get_axis_num(time_dim)
            chunksize = da.data.chunksize if da.chunks is not None else da.shape
            # https://code.mpimet.mpg.de/boards/2/topics/12598
            chunksize = list(chunksize)
            chunksize[time_axis_num] = 1
            chunksize = tuple(chunksize)
            return chunksize
        else:
            return get_chunksizes(da, for_cdo=False, time_dim=time_dim)
    else:
        return da.data.chunksize if da.chunks is not None else da.shape


def get_compress_encoding_nc(
    ds,
    compression="zlib",
    shuffle=True,
    complevel=9,
    for_cdo=False,
    time_dim="time",
    chunks=None,
):
    """Generate encoding for :py:meth:`xarray.Dataset.to_netcdf`.

    Example
    -------
    >>> ds = xr.Dataset({"Tair": (("time", "x", "y"), np.random.rand(36, 20, 10))})
    >>> get_compress_encoding_nc(ds)
    {'Tair': {'zlib': True, 'shuffle': True, 'complevel': 9, 'chunksizes': (36, 20, 10)}}
    >>> get_compress_encoding_nc(ds, for_cdo=True)
    {'Tair': {'zlib': True, 'shuffle': True, 'complevel': 9, 'chunksizes': (1, 20, 10)}}

    See also
    --------
    :py:meth:`xarray.Dataset.to_netcdf`

    """
    enc_checker = xr.backends.netCDF4_._extract_nc4_variable_encoding
    return {
        v: {
            **enc_checker(ds[v]),
            compression: True,
            "shuffle": shuffle,
            "complevel": complevel,
            "chunksizes": get_chunksizes(
                ds[v], for_cdo=for_cdo, time_dim=time_dim, chunks=chunks
            ),
        }
        for v in ds.data_vars
    }


@xr.register_dataset_accessor("to_compressed_netcdf")
class ToCompressed_Netcdf:
    """Save to compressed ``netcdf`` wrapping :py:meth:`xarray.Dataset.to_netcdf` with :py:func:`xbitinfo.save_compressed.get_compress_encoding_nc`.

    Parameters
    ----------
    path : str, path-like or file-like
      Path to which to save this dataset
    compression : str
      Compression library used for encoding. Defaults to ``"zlib"``.
    shuffle : bool
      Netcdf shuffle used for encoding. Defaults to ``True``.
    complevel : int
      Compression level used for encoding.
      Ranges from 2 (little compression, fast) to 9 (strong compression, slow). Defaults to ``7``.
    for_cdo : bool
      If you want to continue working with ``cdo``. If ``True``, sets time chunksize to 1,
      context https://code.mpimet.mpg.de/boards/2/topics/12598. Defaults to ``False``.
    time_dim : str
      Name of the time dimension. Defaults to ``"time"``.
    chunks : str, dict
      How should the data be chunked on disk. None keeps defaults. ``"auto"`` uses ``dask.chunk("auto")``,
      dict individual chunking. Defaults to ``None``.
    kwargs : dict
      Kwargs to be passed to :py:meth:`xarray.Dataset.to_netcdf`

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("rasm")
    >>> path = "compressed_rasm.nc"
    >>> ds.to_compressed_netcdf(path)
    >>> ds.to_compressed_netcdf(path, complevel=4)
    >>> ds.to_compressed_netcdf(path, for_cdo=True)

    See also
    --------
    :py:meth:`xarray.Dataset.to_netcdf`

    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __call__(
        self,
        path,
        compression="zlib",
        shuffle=True,
        complevel=9,
        for_cdo=False,
        time_dim="time",
        chunks=None,
        engine="netcdf4",
        **kwargs,
    ):
        assert engine == "netcdf4", "Only 'netcdf4' engine is currently supported."
        self._obj.to_netcdf(
            path,
            encoding=get_compress_encoding_nc(
                self._obj,
                compression=compression,
                shuffle=shuffle,
                complevel=complevel,
                for_cdo=for_cdo,
                time_dim=time_dim,
                chunks=chunks,
            ),
            engine=engine,
            **kwargs,
        )


def get_compress_encoding_zarr(
    ds,
    compressor=BloscCodec(cname="zstd", shuffle=BloscShuffle.bitshuffle),
    zarr_format="2",
):
    """Generate encoding for :py:meth:`xarray.Dataset.to_zarr`.

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("rasm")
    >>> get_compress_encoding_zarr(ds)  # doctest: +NORMALIZE_WHITESPACE
    {'Tair': {'chunks': None,
              'compressor': BloscCodec(typesize=None,
                                       cname=<BloscCname.zstd: 'zstd'>,
                                       clevel=5,
                                       shuffle=<BloscShuffle.bitshuffle: 'bitshuffle'>,
                                       blocksize=0)}}

    See also
    --------
    :py:meth:`xarray.Dataset.to_zarr`
    """
    encoding = {}
    enc_checker = xr.backends.zarr.extract_zarr_variable_encoding
    if zarr_format == "2":
        compressor_key = "compressor"
    elif zarr_format == "3":
        compressor_key = "compressors"
    if isinstance(compressor, dict):
        default_compressor = BloscCodec(cname="zstd", shuffle=BloscShuffle.bitshuffle)
        encoding = {
            v: {
                **enc_checker(ds[v], zarr_format=zarr_format),
                compressor_key: compressor.get(v, default_compressor),
            }
            for v in ds.data_vars
        }
    else:
        encoding = {
            v: {
                **enc_checker(ds[v], zarr_format=zarr_format),
                compressor_key: compressor,
            }
            for v in ds.data_vars
        }

    # chunks 'auto' can cause issues
    for v, enc in encoding.items():
        chunks = enc.get("chunks", None)
        if chunks is not None:
            if chunks == "auto":
                enc["chunks"] = None

    return encoding


@xr.register_dataset_accessor("to_compressed_zarr")
class ToCompressed_Zarr:
    """Save to compressed ``zarr`` wrapping :py:meth:`xarray.Dataset.to_zarr` with :py:func:`xbitinfo.save_compressed.get_compress_encoding_zarr`.

    Parameters
    ----------
    path : str, path-like or file-like
      Output location of compressed dataset
    compressor : BloscCodec
      Compressor used for encoding. Defaults to zstd with bit-shuffling.
    kwargs : dict
      Arguments to be passed to :py:meth:`xarray.Dataset.to_zarr`

    Example
    -------
    >>> ds = xr.tutorial.load_dataset("rasm")
    >>> path = "compressed_rasm.zarr"
    >>> ds.to_compressed_zarr(path, mode="w")
    >>> ds.to_compressed_zarr(path, compressor=BloscCodec(cname="zlib"), mode="w")
    >>> ds.to_compressed_zarr(
    ...     path, compressor={"Tair": BloscCodec(cname="zstd")}, mode="w"
    ... )

    See also
    --------
    :py:meth:`xarray.Dataset.to_zarr`

    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __call__(
        self,
        path,
        compressor=BloscCodec(cname="zstd", shuffle=BloscShuffle.bitshuffle),
        **kwargs,
    ):
        self._obj.to_zarr(
            path,
            encoding=get_compress_encoding_zarr(
                self._obj,
                compressor=compressor,
            ),
            **kwargs,
        )



"""Functions for initializing the Julia environment and installing deps."""

# Lifted from:
# https://github.com/MilesCranmer/PySR/blob/master/pysr/julia_helpers.py
# https://github.com/MilesCranmer/PySR/blob/master/LICENSE

import os
import warnings
from pathlib import Path

import xbitinfo as xb


def install(julia_project=None, quiet=False):  # pragma: no cover
    """
    Install PyCall.jl and all required dependencies for xbitinfo.

    Also updates the local Julia registry.
    """
    import julia

    julia.install(quiet=quiet)

    julia_project, is_shared = _get_julia_project(julia_project)

    Main = init_julia()
    Main.eval("using Pkg")

    io = "devnull" if quiet else "stderr"
    io_arg = f"io={io}" if is_julia_version_greater_eq(Main, "1.6") else ""

    # Can't pass IO to Julia call as it evaluates to PyObject, so just directly
    # use Main.eval:
    Main.eval(
        f'Pkg.activate("{_escape_filename(julia_project)}", shared = Bool({int(is_shared)}), {io_arg})'
    )
    if is_shared:
        _add_to_julia_project(Main, io_arg)

    Main.eval(f"Pkg.instantiate({io_arg})")
    Main.eval(f"Pkg.precompile({io_arg})")
    if not quiet:
        warnings.warn(
            "It is recommended to restart Python so that the Julia environment is properly initialized."
        )
    already_ran = True
    return already_ran


def import_error_string(julia_project=None):
    s = """
    Required dependencies are not installed or built.  Run the following code in the Python REPL:

        >>> import xbitinfo
        >>> xbitinfo.install()
    """

    if julia_project is not None:
        s += f"""
        Tried to activate project {julia_project} but failed."""

    return s


def _get_julia_project(julia_project):
    if julia_project is None:
        is_shared = True
        julia_project = f"xbitinfo-{xb.__version__}"
    else:
        is_shared = False
        julia_project = Path(julia_project)
    return julia_project, is_shared


def is_julia_version_greater_eq(Main, version="1.6"):
    """Check if Julia version is greater than specified version."""
    return Main.eval(f'VERSION >= v"{version}"')


def init_julia():
    """Initialize julia binary, turning off compiled modules if needed."""
    from julia.core import JuliaInfo, UnsupportedPythonError

    try:
        info = JuliaInfo.load(julia="julia")
    except FileNotFoundError:
        env_path = os.environ["PATH"]
        raise FileNotFoundError(
            f"Julia is not installed in your PATH. Please install Julia and add it to your PATH.\n\nCurrent PATH: {env_path}",
        )

    if not info.is_pycall_built():
        raise ImportError(import_error_string())

    Main = None
    try:
        from julia import Main as _Main

        Main = _Main
    except UnsupportedPythonError:
        # Static python binary, so we turn off pre-compiled modules.
        from julia.core import Julia

        jl = Julia(compiled_modules=False)  # noqa
        from julia import Main as _Main

        Main = _Main

    return Main


def _add_to_julia_project(Main, io_arg):
    Main.bitinformation_spec = Main.PackageSpec(
        name="BitInformation",
        url="https://github.com/milankl/BitInformation.jl",
        rev="v0.6.3",
    )
    Main.statsbase_spec = Main.PackageSpec(
        name="StatsBase",
        url="https://github.com/JuliaStats/StatsBase.jl",
        rev="v0.33.21",
    )
    Main.eval(f"Pkg.add([bitinformation_spec, statsbase_spec], {io_arg})")


def _escape_filename(filename):
    """Turns a file into a string representation with correctly escaped backslashes"""
    str_repr = str(filename)
    str_repr = str_repr.replace("\\", "\\\\")
    return str_repr
